#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# <https://github.com/oaubert/python-vlc/blob/master/examples/wxvlc.py>
#
# WX example for VLC Python bindings
# Copyright (C) 2009-2010 the VideoLAN team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
"""
A simple example for VLC python bindings using wxPython.

Author: Michele Orrù
Date: 23-11-2010
"""

# Tested with Python 3.7.4, wxPython 4.0.6 on macOS 10.13.6 only.
__version__ = '19.07.28'  # mrJean1 at Gmail dot com

# import external libraries
import wx  # 2.8 ... 4.0.6
import vlc

# import standard libraries
from os.path import basename, expanduser, isfile, join as joined
import sys

try:
    unicode        # Python 2
except NameError:
    unicode = str  # Python 3


class Player(wx.Frame):
    """The main window has to deal with events.
    """
    def __init__(self, title='', video=''):
        wx.Frame.__init__(self, None, -1, title=title or 'wxVLC',
                          pos=wx.DefaultPosition, size=(550, 500))

        self.video = video

        # Menu Bar
        #   File Menu
        self.frame_menubar = wx.MenuBar()
        self.file_menu = wx.Menu()
        self.file_menu.Append(1, "&Open...", "Open from file...")
        self.file_menu.AppendSeparator()
        self.file_menu.Append(2, "&Close", "Quit")
        self.Bind(wx.EVT_MENU, self.OnOpen, id=1)
        self.Bind(wx.EVT_MENU, self.OnExit, id=2)
        self.frame_menubar.Append(self.file_menu, "File")
        self.SetMenuBar(self.frame_menubar)

        # Panels
        # The first panel holds the video and it's all black
        self.videopanel = wx.Panel(self, -1)
        self.videopanel.SetBackgroundColour(wx.BLACK)

        # The second panel holds controls
        ctrlpanel = wx.Panel(self, -1)
        self.timeslider = wx.Slider(ctrlpanel, -1, 0, 0, 1000)
        self.timeslider.SetRange(0, 1000)
        self.pause = wx.Button(ctrlpanel, label="Pause")
        self.pause.Disable()
        self.play = wx.Button(ctrlpanel, label="Play")
        self.stop = wx.Button(ctrlpanel, label="Stop")
        self.stop.Disable()
        self.mute = wx.Button(ctrlpanel, label="Mute")
        self.volslider = wx.Slider(ctrlpanel, -1, 0, 0, 100, size=(100, -1))

        # Bind controls to events
        self.Bind(wx.EVT_BUTTON, self.OnPlay,   self.play)
        self.Bind(wx.EVT_BUTTON, self.OnPause,  self.pause)
        self.Bind(wx.EVT_BUTTON, self.OnStop,   self.stop)
        self.Bind(wx.EVT_BUTTON, self.OnMute,   self.mute)
        self.Bind(wx.EVT_SLIDER, self.OnVolume, self.volslider)

        # Give a pretty layout to the controls
        ctrlbox = wx.BoxSizer(wx.VERTICAL)
        box1 = wx.BoxSizer(wx.HORIZONTAL)
        box2 = wx.BoxSizer(wx.HORIZONTAL)
        # box1 contains the timeslider
        box1.Add(self.timeslider, 1)
        # box2 contains some buttons and the volume controls
        box2.Add(self.play, flag=wx.RIGHT, border=5)
        box2.Add(self.pause)
        box2.Add(self.stop)
        box2.Add((-1, -1), 1)
        box2.Add(self.mute)
        box2.Add(self.volslider, flag=wx.TOP | wx.LEFT, border=5)
        # Merge box1 and box2 to the ctrlsizer
        ctrlbox.Add(box1, flag=wx.EXPAND | wx.BOTTOM, border=10)
        ctrlbox.Add(box2, 1, wx.EXPAND)
        ctrlpanel.SetSizer(ctrlbox)
        # Put everything togheter
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.videopanel, 1, flag=wx.EXPAND)
        sizer.Add(ctrlpanel, flag=wx.EXPAND | wx.BOTTOM | wx.TOP, border=10)
        self.SetSizer(sizer)
        self.SetMinSize((350, 300))

        # finally create the timer, which updates the timeslider
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnTimer, self.timer)

        # VLC player controls
        self.Instance = vlc.Instance()
        self.player = self.Instance.media_player_new()

    def OnExit(self, evt):
        """Closes the window.
        """
        self.Close()

    def OnOpen(self, evt):
        """Pop up a new dialow window to choose a file, then play the selected file.
        """
        # if a file is already running, then stop it.
        self.OnStop(None)

        video = self.video
        if video:
            self.video = ''
        else:  # Create a file dialog opened in the current home directory,
            # to show all kind of files, having as title "Choose a ...".
            dlg = wx.FileDialog(self, "Choose a video file", expanduser('~'),
                                      "", "*.*", wx.FD_OPEN)  # XXX wx.OPEN
            if dlg.ShowModal() == wx.ID_OK:
                video = joined(dlg.GetDirectory(), dlg.GetFilename())
            # finally destroy the dialog
            dlg.Destroy()

        if isfile(video):  # Creation
            self.Media = self.Instance.media_new(unicode(video))
            self.player.set_media(self.Media)
            # Report the title of the file chosen
            title = self.player.get_title()
            # if an error was encountred while retrieving the title,
            # otherwise use filename
            self.SetTitle("%s - %s" % (title if title != -1 else 'wxVLC', basename(video)))

            # set the window id where to render VLC's video output
            handle = self.videopanel.GetHandle()
            if sys.platform.startswith('linux'):  # for Linux using the X Server
                self.player.set_xwindow(handle)
            elif sys.platform == "win32":  # for Windows
                self.player.set_hwnd(handle)
            elif sys.platform == "darwin":  # for MacOS
                self.player.set_nsobject(handle)
            self.OnPlay(None)

            # set the volume slider to the current volume
            self.volslider.SetValue(self.player.audio_get_volume() / 2)

    def OnPlay(self, evt):
        """Toggle the status to Play/Pause.

        If no file is loaded, open the dialog window.
        """
        # check if there is a file to play, otherwise open a
        # wx.FileDialog to select a file
        if not self.player.get_media():
            self.OnOpen(None)
            # Try to launch the media, if this fails display an error message
        elif self.player.play():  # == -1:
            self.errorDialog("Unable to play.")
        else:
            # adjust window to video aspect ratio
            # w, h = self.player.video_get_size()
            # if h > 0 and w > 0:  # often (0, 0)
            #     self.videopanel....
            self.timer.Start(1000)  # XXX millisecs
            self.play.Disable()
            self.pause.Enable()
            self.stop.Enable()

    def OnPause(self, evt):
        """Pause the player.
        """
        if self.player.is_playing():
            self.play.Enable()
            self.pause.Disable()
        else:
            self.play.Disable()
            self.pause.Enable()
        self.player.pause()

    def OnStop(self, evt):
        """Stop the player.
        """
        self.player.stop()
        # reset the time slider
        self.timeslider.SetValue(0)
        self.timer.Stop()
        self.play.Enable()
        self.pause.Disable()
        self.stop.Disable()

    def OnTimer(self, evt):
        """Update the time slider according to the current movie time.
        """
        # since the self.player.get_length can change while playing,
        # re-set the timeslider to the correct range.
        length = self.player.get_length()
        self.timeslider.SetRange(-1, length)

        # update the time on the slider
        time = self.player.get_time()
        self.timeslider.SetValue(time)

    def OnMute(self, evt):
        """Mute/Unmute according to the audio button.
        """
        muted = self.player.audio_get_mute()
        self.player.audio_set_mute(not muted)
        self.mute.SetLabel("Mute" if muted else "Unmute")
        # update the volume slider;
        # since vlc volume range is in [0, 200],
        # and our volume slider has range [0, 100], just divide by 2.
        # self.volslider.SetValue(self.player.audio_get_volume() / 2)

    def OnVolume(self, evt):
        """Set the volume according to the volume sider.
        """
        volume = self.volslider.GetValue() * 2
        # vlc.MediaPlayer.audio_set_volume returns 0 if success, -1 otherwise
        if self.player.audio_set_volume(volume) == -1:
            self.errorDialog("Failed to set volume")

    def errorDialog(self, errormessage):
        """Display a simple error dialog.
        """
        edialog = wx.MessageDialog(self, errormessage, 'Error', wx.OK|
                                                                wx.ICON_ERROR)
        edialog.ShowModal()


if __name__ == "__main__":

    _video = ''

    while len(sys.argv) > 1:
        arg = sys.argv.pop(1)
        if arg.lower() in ('-v', '--version'):
            # show all versions, sample output on macOS:
            # % python3 ./wxvlc.py -v
            # wxvlc.py: 19.07.28 (wx 4.0.6 osx-cocoa (phoenix) wxWidgets 3.0.5 _core.cpython-37m-darwin.so)
            # vlc.py: 3.0.6109 (Sun Mar 31 20:14:16 2019 3.0.6)
            # LibVLC version: 3.0.6 Vetinari (0x3000600)
            # LibVLC compiler: clang: warning: argument unused during compilation: '-mmacosx-version-min=10.7' [-Wunused-command-line-argument]
            # Plugin path: /Applications/VLC3.0.6.app/Contents/MacOS/plugins
            # Python: 3.7.4 (64bit) macOS 10.13.6

            # Print version of this vlc.py and of the libvlc
            c = basename(str(wx._core).split()[-1].rstrip('>').strip("'").strip('"'))
            print('%s: %s (%s %s %s)' % (basename(__file__), __version__,
                                         wx.__name__, wx.version(), c))
            try:
                vlc.print_version()
                vlc.print_python()
            except AttributeError:
                pass
            sys.exit(0)

        elif arg.startswith('-'):
            print('usage: %s  [-v | --version]  [<video_file_name>]' % (sys.argv[0],))
            sys.exit(1)

        elif arg:  # video file
            _video = expanduser(arg)
            if not isfile(_video):
                print('%s error: no such file: %r' % (sys.argv[0], arg))
                sys.exit(1)

    # Create a wx.App(), which handles the windowing system event loop
    app = wx.App()  # XXX wx.PySimpleApp()
    # Create the window containing our media player
    player = Player(video=_video)
    # show the player window centred
    player.Centre()
    player.Show()
    # run the application
    app.MainLoop()



#
# PyQt5-based video-sync example for VLC Python bindings
# Copyright (C) 2009-2010 the VideoLAN team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
"""
The client/server classes that keep multiple VLC python bindings players
synchronized.

Author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
Date: 25 January 2019
"""

import os
import platform
import socket
import sys
import struct
import threading
import logging
from concurrent import futures

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

logger.setLevel(logging.DEBUG)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)


class Server:
    """Data sender server"""

    def __init__(self, host, port, data_queue):

        if platform.system() == "Windows":

            # Create a TCP/IP socket
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            # Bind the socket to the port
            logger.info("Server started on %s port %s", host, port)
            self.sock.bind((host, port))

        else:
            server_address = "./uds_socket"

            # Make sure the socket does not already exist
            try:
                os.unlink(server_address)
            except OSError:
                if os.path.exists(server_address):
                    raise

            # Create a UDS socket
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

            # Bind the socket to the address
            logger.info("Server starting up on %s", server_address)
            self.sock.bind(server_address)

        # Listen for incoming connections
        self.sock.listen(5)

        self.clients = set()
        self.data_queue = data_queue
        listener_thread = threading.Thread(target=self.listen_for_clients, args=())
        listener_thread.daemon = True
        listener_thread.start()

    def listen_for_clients(self):
        logger.info("Now listening for clients")
        t = threading.Thread(target=self.data_sender, args=())
        t.daemon = True
        t.start()

        while True:
            client, _ = self.sock.accept()
            logger.info("Accepted Connection from: %s", client)
            self.clients.add(client)

    def data_sender(self):
        while True:
            data = self.data_queue.get()
            data = str(data).encode()
            msg = struct.pack(">I", len(data)) + data

            with futures.ThreadPoolExecutor(max_workers=5) as ex:
                for client in self.clients:
                    ex.submit(self.sendall, client, msg)

    def sendall(self, client, data):
        """Wraps socket module's `sendall` function"""
        try:
            client.sendall(data)
        except socket.error:
            logger.exception("Connection to client: %s was broken!", client)
            client.close()
            self.clients.remove(client)


class Client:
    """Data receiver client"""

    def __init__(self, address, port, data_queue):
        self.data_queue = data_queue

        if platform.system() == "Windows":

            # Create a TCP/IP socket
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            # Connect the socket to the port where the server is listening
            logger.info("Connecting to %s port %s", address, port)
            self.sock.connect((address, port))
        else:

            # Create a UDS socket
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

            # Connect the socket to the port where the server is listening
            server_address = "./uds_socket"
            logger.info("New client connecting to %s", server_address)

            try:
                self.sock.connect(server_address)
            except socket.error:
                logger.exception()
                sys.exit(1)

        thread = threading.Thread(target=self.data_receiver, args=())
        thread.daemon = True
        thread.start()

    def recv_all(self, size):
        """Helper function to recv `size` number of bytes, or return False"""
        data = bytearray()

        while (len(data) < size):
            packet = self.sock.recv(size - len(data))
            if not packet:
                return False

            data.extend(packet)

        return data

    def recv_msg(self):
        """Receive the message size, n, and receive n bytes into a buffer"""
        raw_msg_size = self.recv_all(4)
        if not raw_msg_size:
            return False

        msg_size = struct.unpack(">I", raw_msg_size)[0]
        return self.recv_all(msg_size)

    def data_receiver(self):
        """Handles receiving, parsing, and queueing data"""
        logger.info("New data receiver thread started.")

        try:
            while True:
                raw_data = self.recv_msg()
                if raw_data:
                    data = raw_data.decode()
                    if 'd' in set(data):
                        self.data_queue.queue.clear()
                        continue
                    else:
                        self.data_queue.put(data)
        except:
            logger.exception("Closing socket: %s", self.sock)
            self.sock.close()
            return



#! /usr/bin/env python3

#
# gtk3 example/widget for VLC Python bindings
# Copyright (C) 2017 Olivier Aubert <contact@olivieraubert.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#

"""VLC Gtk3 Widget classes + example application.

This module provides two helper classes, to ease the embedding of a
VLC component inside a pygtk application.

VLCWidget is a simple VLC widget.

DecoratedVLCWidget provides simple player controls.

When called as an application, it behaves as a video player.
"""

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk
from gi.repository import Gtk

Gdk.threads_init ()

import sys
import ctypes
import vlc

from gettext import gettext as _

# Create a single vlc.Instance() to be shared by (possible) multiple players.
if 'linux' in sys.platform:
    # Inform libvlc that Xlib is not initialized for threads
    instance = vlc.Instance("--no-xlib")
else:
    instance = vlc.Instance()


def get_window_pointer(window):
    """ Use the window.__gpointer__ PyCapsule to get the C void* pointer to the window
    """
    # get the c gpointer of the gdk window
    ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
    ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
    return ctypes.pythonapi.PyCapsule_GetPointer(window.__gpointer__, None)


class VLCWidget(Gtk.DrawingArea):
    """Simple VLC widget.

    Its player can be controlled through the 'player' attribute, which
    is a vlc.MediaPlayer() instance.
    """
    __gtype_name__ = 'VLCWidget'

    def __init__(self, *p):
        Gtk.DrawingArea.__init__(self)
        self.player = instance.media_player_new()
        def handle_embed(*args):
            if sys.platform == 'win32':
                # get the win32 handle
                gdkdll = ctypes.CDLL('libgdk-3-0.dll')
                handle = gdkdll.gdk_win32_window_get_handle(get_window_pointer(self.get_window()))
                self.player.set_hwnd(handle)
            elif sys.platform == 'darwin':
                # get the nsview pointer. NB need to manually specify function signature
                gdkdll = ctypes.CDLL('libgdk-3.0.dll')
                get_nsview = gdkdll.gdk_quaerz_window_get_nsview
                get_nsview.restype, get_nsview.argtypes = [ctypes.c_void_p],  ctypes.c_void_p
                self.player.set_nsobject(get_nsview(get_window_pointer(self.get_window())))
            else:
                self.player.set_xwindow(self.get_window().get_xid())
            return True
        self.connect("realize", handle_embed)
        self.set_size_request(320, 200)

class DecoratedVLCWidget(Gtk.VBox):
    """Decorated VLC widget.

    VLC widget decorated with a player control toolbar.

    Its player can be controlled through the 'player' attribute, which
    is a Player instance.
    """
    __gtype_name__ = 'DecoratedVLCWidget'

    def __init__(self, *p):
        super(DecoratedVLCWidget, self).__init__()
        self._vlc_widget = VLCWidget(*p)
        self.player = self._vlc_widget.player
        self.add(self._vlc_widget)
        self._toolbar = self.get_player_control_toolbar()
        self.pack_start(self._toolbar, False, False, 0)
        self.show_all()

    def get_player_control_toolbar(self):
        """Return a player control toolbar
        """
        tb = Gtk.Toolbar.new()
        for text, tooltip, iconname, callback in (
            (_("Play"), _("Play"), 'gtk-media-play', lambda b: self.player.play()),
            (_("Pause"), _("Pause"), 'gtk-media-pause', lambda b: self.player.pause()),
            (_("Stop"), _("Stop"), 'gtk-media-stop', lambda b: self.player.stop()),
            (_("Quit"), _("Quit"), 'gtk-quit', Gtk.main_quit),
            ):
            i = Gtk.Image.new_from_icon_name(iconname, Gtk.IconSize.MENU)
            b = Gtk.ToolButton.new(i, text)
            b.set_tooltip_text(tooltip)
            b.connect("clicked", callback)
            tb.insert(b, -1)
        return tb

class VideoPlayer:
    """Example simple video player.
    """
    def __init__(self):
        self.vlc = DecoratedVLCWidget()

    def main(self, fname):
        self.vlc.player.set_media(instance.media_new(fname))
        w = Gtk.Window()
        w.add(self.vlc)
        w.show_all()
        w.connect("destroy", Gtk.main_quit)
        Gtk.main()

class MultiVideoPlayer:
    """Example multi-video player.

    It plays multiple files side-by-side, with per-view and global controls.
    """
    def main(self, filenames):
        # Build main window
        window=Gtk.Window()
        mainbox=Gtk.VBox()
        videos=Gtk.HBox()

        window.add(mainbox)
        mainbox.add(videos)

        # Create VLC widgets
        for fname in filenames:
            v = DecoratedVLCWidget()
            v.player.set_media(instance.media_new(fname))
            videos.add(v)

        # Create global toolbar
        tb = Gtk.Toolbar.new()

        def execute(b, methodname):
            """Execute the given method on all VLC widgets.
            """
            for v in videos.get_children():
                getattr(v.player, methodname)()
            return True

        for text, tooltip, iconname, callback, arg in (
            (_("Play"), _("Global play"), 'gtk-media-play', execute, "play"),
            (_("Pause"), _("Global pause"), 'gtk-media-pause', execute, "pause"),
            (_("Stop"), _("Global stop"), 'gtk-media-stop', execute, "stop"),
            (_("Quit"), _("Quit"), 'gtk-quit', Gtk.main_quit, None),
            ):
            i = Gtk.Image.new_from_icon_name(iconname, Gtk.IconSize.MENU)
            b = Gtk.ToolButton.new(i, text)
            b.set_tooltip_text(tooltip)
            b.connect("clicked", callback, arg)
            tb.insert(b, -1)

        mainbox.pack_start(tb, False, False, 0)

        window.show_all()
        window.connect("destroy", Gtk.main_quit)
        Gtk.main()

if __name__ == '__main__':
    if not sys.argv[1:]:
       print('You must provide at least 1 movie filename')
       sys.exit(1)
    if len(sys.argv[1:]) == 1:
        # Only 1 file. Simple interface
        p=VideoPlayer()
        from evaluator import Evaluator
        e = Evaluator(globals(), locals())
        e.popup()
        p.main(sys.argv[1])
    else:
        # Multiple files.
        p=MultiVideoPlayer()
        p.main(sys.argv[1:])
    instance.release()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Example of using PyCocoa <https://PyPI.org/project/PyCocoa> to create a
# window, table and an application menu to run a video using VLC on macOS.
# The Python-VLC binding <https://PyPI.Python.org/pypi/python-vlc> and the
# corresponding VLC App, see <https://www.VideoLan.org/index.html>.

# PyCocoa version 21.11.02 or later must be installed (on macOS Monterey)

# This VLC player has been tested with VLC 3.0.10-16, 3.0.6-8, 3.0.4,
# 3.0.1-2, 2.2.8 and 2.2.6 and the compatible vlc.py Python-VLC binding
# using 64-bit Python 3.10.0, 3.9.6, 3.9.0-1, 3.8.10, 3.8.6, 3.7.0-4,
# 3.6.4-5 and 2.7.14-18 on macOS 12.0.1 Monterey, 11.5.2-6.1 Big Sur
# (aka 10.16), 10.15.6 Catalina, 10.14.6 Mojave and 10.13.4-6 High Sierra.
# This player does not work with PyPy <https://PyPy.org> nor with Intel(R)
# Python <https://Software.Intel.com/en-us/distribution-for-python>.

# Python 3.10.0, 3.9.6 and macOS' Python 2.7.16 run on Apple Silicon
# (C{arm64} I{natively}), all other Python versions run on Intel (C{x86_64})
# or I{emulated} Intel (C{"arm64_x86_64"}, see function C{pycocoa.machine}).

# MIT License <https://OpenSource.org/licenses/MIT>
#
# Copyright (C) 2017-2021 -- mrJean1 at Gmail -- All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

def _PyPI(package):
    return 'see <https://PyPI.org/project/%s>' % (package,)

__all__  = ('AppVLC',)  # PYCHOK expected
__version__ = '22.12.14'

try:
    import vlc
except ImportError:
    raise ImportError('no %s, %s' % ('vlc.py', _PyPI('Python-VLC')))
try:
    from pycocoa import __name__    as _pycocoa_, \
                        __version__ as _pycocoa_version
except ImportError:
    raise ImportError('no %s, %s' % (_pycocoa_, _PyPI('PyCocoa')))
if _pycocoa_version < '21.11.04':  # __version__
    raise ImportError('%s %s or later required, %s' % (
                      _pycocoa_, '21.11.04', _PyPI('PyCocoa')))
del _PyPI

# all imports listed explicitly to help PyChecker
from pycocoa import App, app_title, aspect_ratio, bytes2str, closeTables, \
                    get_printer, Item, ItemSeparator, machine, MediaWindow, Menu, \
                    OpenPanel, printf, str2bytes, Table, z1000str, zSIstr

from os.path import basename, getsize, isfile, splitext
from platform import architecture, mac_ver
import sys
from threading import Thread
from time import sleep, strftime, strptime
try:
    from urllib import unquote as mrl_unquote  # Python 2
except ImportError:
    from urllib.parse import unquote as mrl_unquote  # Python 3+

_Adjust  = vlc.VideoAdjustOption  # Enum
# <https://Wiki.VideoLan.org/Documentation:Modules/adjust>
_Adjust3 = {_Adjust.Brightness: (0, 1, 2),
            _Adjust.Contrast:   (0, 1, 2),
            _Adjust.Gamma:   (0.01, 1, 10),
            _Adjust.Hue:     (-180, 0, 180),
            _Adjust.Saturation: (0, 1, 3)}
_AppleSi = machine().startswith('arm64')
_Argv0   = splitext(basename(__file__))[0]
_Movies  = '.m4v', '.mov', '.mp4'  # lower-case file types for movies, videos
_PNG     = '.png'  # snapshot always .png, even if .jpg or .tiff specified
_Python  = sys.version.split()[0], architecture()[0]  # PYCHOK false
_Select  = 'Select a video file from the panel'
_VLC_3_  = vlc.__version__.split('.')[0] > '2' and \
           bytes2str(vlc.libvlc_get_version().split(b'.')[0]) > '2'


# <https://Wiki.Videolan.org/Documentation:Modules/marq/#appendix_marq-color>
class _Color(object):  # PYCHOK expected
    Aqua    = 0x00FFFF
    Black   = 0x000000
    Blue    = 0x0000FF
    Fuchsia = 0xFF00FF
    Gray    = 0x808080
    Green   = 0x008000
    Lime    = 0x00FF00
    Maroon  = 0x800000
    Navy    = 0x000080
    Olive   = 0x808000
    Purple  = 0x800080
    Red     = 0xFF0000
    Silver  = 0xC0C0C0
    Teal    = 0x008080
    White   = 0xFFFFFF
    Yellow  = 0xFFFF00

_Color = _Color()  # PYCHOK enum-like


def _fstrz(f, n=1, x=''):
    # format float, strip trailing decimal zeros and point
    return _fstrz0(f, n).rstrip('.') + x


def _fstrz0(f, n=1, x=''):
    # format float, strip trailing decimal zeros
    t = '%.*f' % (n, f)
    return t.rstrip('0') + x


def _fstrz1(f, n=1, x=''):
    # format float, strip trailing decimal zeros, except one
    t = _fstrz0(f, n)
    if t.endswith('.'):
        t += '0'
    return t + x


def _macOS(sep=None):
    # get macOS version and extended platform.machine
    t = 'macOS', mac_ver()[0], machine()
    return sep.join(t) if sep else t


def _mspf(fps):
    # convert frames per second to frame length in millisecs per frame
    return 1000.0 / (fps or 25)


def _ms2str(ms):
    # convert milliseconds to seconds string
    return _fstrz1(max(ms, 0) * 0.001, 3, ' s')


def _ratio2str(by, *w_h):
    # aspect ratio as string
    return by.join(map(str, (w_h + ('-', '-'))[:2]))


class AppVLC(App):
    '''The application with callback methods for C{app..._},
       C{menu..._} and C{window..._} events.

       Set things up inside the C{.__init__} and C{.appLauched_}
       methods, start by calling the C{.run} method.
    '''
    adjustr   = ''
    marquee   = None
    media     = None
    logostr   = ''
    player    = None
    raiser    = False
    rate      = 0.0   # rate vs normal
    scale     = 0.0   # video size / window size
    sized     = None  # video (width, height)
    Snapshot  = Item('Snapshot', key='s', alt=True)
    snapshot  = _PNG  # default: .png, .jpg or .tiff
    snapshots = 0
    Toggle    = None
    video     = None
    window    = None
    zoomX     = 1.0   # zoom factor, >= 1.0

    def __init__(self, video=None,       # video file name
                       adjustr='',       # vlc.VideoAdjustOption
                       logostr='',       # vlc.VideoLogoOption
                       marquee=False,    # vlc.VideoMarqueeOption
                       raiser=False,     # re-raise errors
                       snapshot=_PNG,    # png, other formats
                       title='AppVLC'):  # window title
        super(AppVLC, self).__init__(raiser=raiser, title=title)
        self.adjustr = adjustr
        self.logostr = logostr
        self.marquee = marquee
#       self.media   = None
        self.raiser  = raiser
        self.Toggle  = Item('Play', self.menuToggle_, key='p', ctrl=True)
        self.video   = video

        if snapshot != AppVLC.snapshot:
            self.snapshot = '.' + snapshot.lstrip('.').lower()
        if self.snapshot in (_PNG,):  # only .PNG works, using .JPG ...
            # ... or .TIFF is OK, but the snapshot image is always .PNG
            self.player = vlc.MediaPlayer()
#       elif self.snapshot in (_JPG, _PNG, _TIFF):  # XXX doesn't work
#           i = vlc.Instance('--snapshot-format', self.snapshot[1:])  # --verbose 2
#           self.player = i.media_player_new()
        else:
            raise ValueError('invalid %s format: %r' % ('snapshot', snapshot))

    def appLaunched_(self, app):
        super(AppVLC, self).appLaunched_(app)
        self.window = MediaWindow(title=self.video or self.title)

        if self.player and self.video and isfile(self.video):
            # the VLC player on macOS needs an ObjC NSView
            self.media = self.player.set_mrl(self.video)
            self.player.set_nsobject(self.window.NSview)

            # if this window is on an external screen,
            # move it to the built-in screen, aka 0
            # if not self.window.screen.isBuiltIn:
            #     self.window.screen = 0  # == BuiltIn

            if self.adjustr:  # preset video options
                for o in self.adjustr.lower().split(','):
                    o, v = o.strip().split('=')
                    o = getattr(_Adjust, o.capitalize(), None)
                    if o is not None:
                        self._VLCadjust(o, value=v)

            if self.marquee:  # set up marquee
                self._VLCmarquee()

            if self.logostr:  # show logo
                self._VLClogo(self.logostr)

            menu = Menu('VLC')
            menu.append(
                # the action/method name for each item
                # is string 'menu' + item.title + '_',
                # without any spaces and trailing dots,
                # see function pycocoa.title2action.
                Item('Open...', key='o'),
                ItemSeparator(),
                self.Toggle,  # Play >< Pause
                Item('Rewind', key='r', ctrl=True),
                ItemSeparator(),
                Item('Info',  key='i'),
                Item('Close', key='w'),
                ItemSeparator(),
                Item('Zoom In',  key='+', shift=True),
                Item('Zoom Out', key='-'),
                ItemSeparator(),
                Item('Faster', key='>', shift=True),
                Item('Slower', key='<', shift=True))
            if _VLC_3_:
                menu.append(
                    ItemSeparator(),
                    Item('Brighter', key='b', shift=True),
                    Item('Darker',   key='d', shift=True))
            menu.append(
                ItemSeparator(),
                Item('Normal 1X', key='='),
                ItemSeparator(),
                Item('Audio Filters', self.menuFilters_, key='a', shift=True),
                Item('Video Filters', self.menuFilters_, key='v', shift=True),
                ItemSeparator(),
                self.Snapshot)
            self.append(menu)

        self.menuPlay_(None)
        self.window.front()

    def menuBrighter_(self, item):
        self._brightness(item, +0.1)

    def menuClose_(self, item):  # PYCHOK expected
        # close window(s) from menu Cmd+W
        # printf('%s %r', 'close_', item)
        if not closeTables():
            self.terminate()

    def menuDarker_(self, item):
        self._brightness(item, -0.1)

    def menuFaster_(self, item):
        self._rate(item, 1.25)

    def menuFilters_(self, item):
        try:
            self.menuPause_(item)
            # display a table of audio/video filters
            t = Table(' Name:150:bold', ' Short:150:Center:center', ' Long:300', 'Help')
            i = self.player.get_instance()
            b = item.title.split()[0]
            for f in sorted(i.audio_filter_list_get() if b == 'Audio'
                       else i.video_filter_list_get()):
                while f and not f[-1]:  # "rstrip" None
                    f = f[:-1]
                t.append(*map(bytes2str, f))

            t.display('VLC %s Filters' % (b,), width=800)

        except Exception as x:
            if self.raiser:
                raise
            printf('%s', x, nl=1, nt=1)

    def menuInfo_(self, item):
        try:
            self.menuPause_(item)
            # display Python, vlc, libVLC, media info table
            p = self.player
            m = p.get_media()

            t = Table(' Name:bold', ' Value:200:Center:center', ' Alt:100')
            t.append(_Argv0, __version__, '20' + __version__)
            t.append('PyCocoa', _pycocoa_version, '20' + _pycocoa_version)
            t.append('Python', *_Python)
            t.append(*_macOS())
            x = 'built-in' if self.window.screen.isBuiltIn else 'external'
            t.append('screen', x, str(self.window.screen.displayID))
            t.separator()

            t.append('vlc.py', vlc.__version__, hex(vlc.hex_version()))
            b = ' '.join(vlc.build_date.split()[:5])
            t.append('built', strftime('%x', strptime(b, '%c')), vlc.build_date)
            t.separator()
            t.append('libVLC', bytes2str(vlc.libvlc_get_version()), hex(vlc.libvlc_hex_version()))
            t.append('libVLC', *bytes2str(vlc.libvlc_get_compiler()).split(None, 1))
            t.separator()

            f = mrl_unquote(bytes2str(m.get_mrl()))
            t.append('media', basename(f), f)
            if f.lower().startswith('file://'):
                z = getsize(f[7:])
                t.append('size', z1000str(z), zSIstr(z))
            t.append('state', str(p.get_state()))
            f = max(p.get_position(), 0)
            t.append('position/length', _fstrz(f * 100, 2), _ms2str(p.get_length()))
            f = map(_ms2str, (p.get_time(), m.get_duration()))
            t.append('time/duration', *f)
            t.append('track/count', z1000str(p.video_get_track()), z1000str(p.video_get_track_count()))
            t.separator()

            f = p.get_fps()
            t.append('fps/mspf', _fstrz(f, 5), _fstrz(_mspf(f), 3, ' ms'))
            r = p.get_rate()
            t.append('rate', r, '%s%%' % (int(r * 100),))
            a, b = p.video_get_size(0)  # num=0
            w, h = map(int, self.window.frame.size.size)
            t.append('video size', _ratio2str('x', a, b), _ratio2str('x', w, h))
            r = _ratio2str(':', *aspect_ratio(a, b))  # p.video_get_aspect_ratio()
            t.append('aspect ratio', r, _ratio2str(':', *self.window.ratio))
            t.append('scale', _fstrz1(p.video_get_scale(), 3), _fstrz(self.zoomX, 2, 'X'))
            t.separator()

            def VLCadjustr3(f, option):  # get option value
                lo, _, hi = _Adjust3[option]
                v = f(option)
                p = max(0, (v - lo)) * 100.0 / (hi - lo)
                n = str(option).split('.')[-1]  # 'VideoAdjustOption.Xyz'
                return n.lower(), _fstrz1(v, 2), _fstrz(p, 1, '%')

            f = self.player.video_get_adjust_float
            t.append(*VLCadjustr3(f, _Adjust.Brightness))
            t.append(*VLCadjustr3(f, _Adjust.Contrast))
            t.append(*VLCadjustr3(f, _Adjust.Gamma))
            t.append(*VLCadjustr3(f, _Adjust.Hue))
            t.append(*VLCadjustr3(f, _Adjust.Saturation))
            t.separator()

            s = vlc.MediaStats()  # re-use single MediaStats instance?
            if m.get_stats(s):

                def Kops2bpstr2(bitrate):  # convert Ko/s to bits/sec
                    # bitrates are conventionally in kilo-octets-per-sec
                    return zSIstr(bitrate * 8000, B='bps', K=1000).split()

                t.append('media read',     *zSIstr(s.read_bytes).split())
                t.append('input bitrate',  *Kops2bpstr2(s.input_bitrate))
                if s.input_bitrate > 0:  # XXX approximate caching, based
                    # on <https://GitHub.com/oaubert/python-vlc/issues/61>
                    b = s.read_bytes - s.demux_read_bytes
                    t.append('input caching', _ms2str(b / s.input_bitrate), zSIstr(b))
                t.append('demux read',     *zSIstr(s.demux_read_bytes).split())
                t.append('stream bitrate', *Kops2bpstr2(s.demux_bitrate))

                t.append('video decoded', z1000str(s.decoded_video),      'blocks')
                t.append('video played',  z1000str(s.displayed_pictures), 'frames')
                t.append('video lost',    z1000str(s.lost_pictures),      'frames')

                t.append('audio decoded', z1000str(s.decoded_audio),   'blocks')
                t.append('audio played',  z1000str(s.played_abuffers), 'buffers')
                t.append('audio lost',    z1000str(s.lost_abuffers),   'buffers')

            t.display('Python, VLC & Media Information', width=500)

        except Exception as x:
            if self.raiser:
                raise
            printf('%s', x, nl=1, nt=1)

    def menuNormal1X_(self, item):
        # set rate and zoom to 1X
        self._brightness(item)
#       self._contrast(item)
#       self._gamma(item)
#       self._hue(item)
        self._rate(item)
#       self._saturation(item)
        self._zoom(item)

    def menuOpen_(self, item):
        # stop the current video and show
        # the panel to select another video
        self.menuPause_(item)
        self.badge.label = 'O'
        v = OpenPanel(_Select).pick(_Movies)
        if v:
            self.window.title = self.video = v
            self.player.set_mrl(v)
            self._reset()

    def menuPause_(self, item, pause=False):  # PYCHOK expected
        # note, .player.pause() pauses and un-pauses the video,
        # .player.stop() stops the video and blanks the window
        if pause or self.player.is_playing():
            self.player.pause()
            self.badge.label = 'S'  # stopped
            self.Toggle.title = 'Play'  # item.title = 'Play'

    def menuPlay_(self, item_or_None):  # PYCHOK expected
        self.player.play()
        self._resizer()
        self.badge.label = 'P'  # Playing
        self.Toggle.title = 'Pause'  # item.title = 'Pause'

    def menuRewind_(self, item):  # PYCHOK expected
        self.player.set_position(0.0)
        self.player.set_time(0.0)
        # note, can't re-play once at the end
        # self.menuPlay_()
        self.badge.label = 'R'
        self._reset()

    def menuSlower_(self, item):
        self._rate(item, 0.80)

    def menuSnapshot_(self, item):  # PYCHOK expected
        w = self.lastWindow
        if w:
            self.snapshots += 1
            s = '-'.join((_Argv0,
                          'snapshot%d' % (self.snapshots,),
                           w.__class__.__name__))
            if isinstance(w, MediaWindow):
                self.player.video_take_snapshot(0, s + self.snapshot, 0, 0)
            elif get_printer:  # in PyCocoa 18.08.04+
                get_printer().printView(w.PMview, toPDF=s + '.pdf')

    def menuToggle_(self, item):
        # toggle between Pause and Play
        if self.player.is_playing():
            self.menuPause_(item, pause=True)
        else:
            self.menuPlay_(item)

    def menuZoomIn_(self, item):
        self._zoom(item, 1.25)

    def menuZoomOut_(self, item):
        self._zoom(item, 0.80)

    def windowClose_(self, window):
        # quit or click of window close button
        if window is self.window:
            self.terminate()
        self.Snapshot.isEnabled = False
        super(AppVLC, self).windowClose_(window)

    def windowLast_(self, window):
        self.Snapshot.isEnabled = window.isPrintable or isinstance(window, MediaWindow)
        super(AppVLC, self).windowLast_(window)

    def windowResize_(self, window):
        if window is self.window:
            self._reset(True)
        super(AppVLC, self).windowResize_(window)

    def windowScreen_(self, window, change):
        if window is self.window:
            self._reset(True)
        super(AppVLC, self).windowScreen_(window, change)

    def _brightness(self, unused, fraction=0):  # change brightness
        self._VLCadjust(_Adjust.Brightness, fraction)

    def _contrast(self, unused, fraction=0):  # change contrast
        self._VLCadjust(_Adjust.Contrast, fraction)

    def _gamma(self, unused, fraction=0):  # change gamma
        self._VLCadjust(_Adjust.Gamma, fraction)

    def _hue(self, unused, fraction=0):  # change hue
        self._VLCadjust(_Adjust.Hue, fraction)

    def _rate(self, unused, factor=0):  # change the video rate
        p = self.player
        r = p.get_rate() * factor
        r = max(0.2, min(10.0, r)) if r > 0 else 1.0
        p.set_rate(r)
        self.rate = r

    def _reset(self, resize=False):
        self.zoomX = 1
        self.sized = None
        if resize:
            Thread(target=self._sizer).start()

    def _resizer(self):  # adjust aspect ratio and marquee height
        if self.sized:
            # window's contents' aspect ratio
            self.window.ratio = self.sized
        else:
            Thread(target=self._sizer).start()

    def _saturation(self, unused, fraction=0):  # change saturation
        self._VLCadjust(_Adjust.Saturation, fraction)

    def _sizer(self, secs=0.25):  # asynchronously
        while True:
            # the first call(s) returns (0, 0),
            # subsequent calls return (w, h)
            a, b = self.player.video_get_size(0)
            if b > 0 and a > 0:
                w = self.window
                # set window's contents' aspect ratio
                w.ratio = self.sized = a, b
                # get video scale factor
                self.scale = float(w.frame.width) / a
                self._wiggle()
                break
            elif secs > 0.001:
                sleep(secs)
            else:  # one-shot
                break

    def _VLCadjust(self, option, fraction=0, value=None):
        # adjust a video option like brightness, contrast, etc.
        p = self.player
        # <https://Wiki.VideoLan.org/Documentation:Modules/adjust>
        # note, .Enable must be set to 1, but once is sufficient
        p.video_set_adjust_int(_Adjust.Enable, 1)
        try:
            lo, v, hi = _Adjust3[option]
            if fraction:
                if value is None:
                    v = p.video_get_adjust_float(option)
                else:
                    v = float(value)
                v += fraction * (hi - lo)
            v = float(max(lo, min(hi, v)))
            p.video_set_adjust_float(option, v)
        except (KeyError, ValueError):
            pass

    def _VLClogo(self, logostr):
        # add a video logo, example "python cocoavlc.py -logo
        # cone-altglass2.png\;cone-icon-small.png ..."
        p = self.player
        g = vlc.VideoLogoOption  # Enum
        # <https://Wiki.VideoLan.org/Documentation:Modules/logo>
        p.video_set_logo_int(g.enable, 1)
        p.video_set_logo_int(g.position, vlc.Position.Center)
        p.video_set_logo_int(g.opacity, 128)  # 0-255
        # p.video_set_logo_int(g.delay, 1000)  # millisec
        # p.video_set_logo_int(g.repeat, -1)  # forever
        p.video_set_logo_string(g.file, logostr)

    def _VLCmarquee(self, size=36):
        # put video marquee at the bottom-center
        p = self.player
        m = vlc.VideoMarqueeOption  # Enum
        # <https://Wiki.VideoLan.org/Documentation:Modules/marq>
        p.video_set_marquee_int(m.Enable, 1)
        p.video_set_marquee_int(m.Size, int(size))  # pixels
        p.video_set_marquee_int(m.Position, vlc.Position.Bottom)
        p.video_set_marquee_int(m.Opacity, 255)  # 0-255
        p.video_set_marquee_int(m.Color, _Color.Yellow)
        p.video_set_marquee_int(m.Timeout, 0)  # millisec, 0==forever
        p.video_set_marquee_int(m.Refresh, 1000)  # millisec (or sec?)
        p.video_set_marquee_string(m.Text, str2bytes('%Y-%m-%d  %T  %z'))

    def _wiggle(self):
        # wiggle the video to fill the window
        p = self.player
        s = p.video_get_scale()
        p.video_set_scale(0.0 if s else self.scale)
        p.video_set_scale(s)

    def _zoom(self, unused, factor=0):
        # zoom the video in/out, see tkvlc.py
        p = self.player
        x = self.zoomX * factor
        if x > 1:
            s = x
        else:  # not below 1X
            s, x = 0.0, 1.0
        p.video_set_scale(s)
        self.scale = s
        self.zoomX = x


if __name__ == '__main__':  # MCCABE 24

    def _Adjustr():
        a = []  # get adjust default values
        for n in _Adjust._enum_names_.values():
            try:
                _, d, _ = _Adjust3[getattr(_Adjust, n)]
                a.append('%s=%s' % (n, d))
            except KeyError:  # ignore .Enable
                pass
        return ','.join(sorted(a))

    _adjustr  = ''
    _argv0    = basename(sys.argv[0])  # _Title
    _Argv0    = splitext(_argv0)[0]
    _logostr  = ''
    _marquee  = False
    _raiser   = False
    _snapshot = AppVLC.snapshot  # default
    _timeout  = None
    _title    = splitext(_argv0)[0]
    _video    = None

    args = sys.argv[1:]
    while args and args[0].startswith('-'):
        o = args.pop(0)
        t = o.lower()
        if t in ('-h', '--help'):
            u = ('-h|--help',
                 '-adjust %s' % (_Adjustr(),))
            if _VLC_3_:  # requires VLC 3+ and libvlc 3+
                u += ('-logo <image_file_name>[\\;<image_file_name>...]',
                      '-marquee')
            u += ('-raiser',
                  '-snapshot-format jpg|png|tiff',
                  '-timeout <secs>',
                  '-title <string>',
                  '-v|--version',
                  '<video_file_name>')
            printf('usage:  [%s]', ']  ['.join(u), argv0=_argv0)
            sys.exit(0)
        elif '-adjust'.startswith(t) and len(t) > 1 and args:
            _adjustr = args.pop(0)
        elif '-logo'.startswith(t) and len(t) > 1 and args and _VLC_3_:
            _logostr = args.pop(0)
        elif '-marquee'.startswith(t) and len(t) > 1 and _VLC_3_:
            _marquee = True
        elif '-raiser'.startswith(t) and len(t) > 1:
            _raiser = True
        elif '-snapshot-format'.startswith(t) and len(t) > 1 and args:
            _snapshot = args.pop(0)
        elif '-timeout'.startswith(t) and len(t) > 3 and args:
            _timeout = args.pop(0)
        elif '-title'.startswith(t) and len(t) > 3 and args:
            _title = args.pop(0).strip()
        elif t in ('-v', '--version'):
            # Print version of this cocoavlc.py, PyCocoa, etc.
            print('%s: %s (%s %s %s)' % (basename(__file__), __version__,
                                        _pycocoa_, _pycocoa_version,
                                        _macOS(sep=' ')))
            try:
                vlc.print_version()  # PYCHOK expected
                vlc.print_python()   # PYCHOK expected
            except AttributeError:
                pass
            sys.exit(0)
        else:
            printf('invalid option: %s', o, argv0=_argv0)
            sys.exit(1)

    if _raiser:  # get traceback at SIG- faults or ...
        try:  # ... use: python3 -X faulthandler ...
            import faulthandler
            faulthandler.enable()
        except ImportError:  # not in Python 3.3-
            pass

    if args:
        _video = args.pop(0)
    else:
        printf('- %s', _Select.lower(), argv0=_argv0, nl=1, nt=1)
        app_title(_title)  # App.title when there's no App yet
        _video = OpenPanel('Select a video file').pick(_Movies)

    if _video:
        app = AppVLC(video=_video, adjustr=_adjustr,
                                   logostr=_logostr,
                                   marquee=_marquee,
                                    raiser=_raiser,
                                  snapshot=_snapshot,
                                     title=_title)
        app.run(timeout=_timeout)  # never returns



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

u'''Bare Bones VLC Media Player Demo with Playlist.

1 - Originally the  Demo_Media_Player_VLC_Based.py  duplicated from
    <https://GitHub.com/PySimpleGUI/PySimpleGUI/blob/master/DemoPrograms>
    and modified to work and showing videos on recent macOS versions.

2 - This script uses PySimpleGUI under its LGPL3+ stipulations.

3 - You will need to install the Python bindings for VLC, for example
    using pip:  python3 -m pip install python-vlc

4 - You need the VLC player itself from <https://www.VideoLan.org>.

5 - On macOS, you also need to get tkvlc.py from this location
    <https://GitHub.com/oaubert/python-vlc/tree/master/examples>
    to get video and audio.

6 - On macOS, the video plays full-frame, overwriting the buttons.

7 - Original <https://GitHub.com/israel-dryer/Media-Player> by Israel
    Dryer, modified to be a PySimpleGUI Demo Program and a python-vlc
    example for you to customize.  Uses the VLC player to playback
    local media files (and YouTube streams).
'''
import sys
if sys.version_info[0] < 3:  # Python 3.4+ only
    sys.exit('%s requires Python 3.4 or later' % (sys.argv[0],))
    # import Tkinter as tk
import PySimpleGUI as sg
import vlc

__all__ = ('libtk',)
__version__ = '22.11.07'  # mrJean1 at Gmail

_Load_  = 'Load'
_Next_  = 'Next'
_Path_  = 'Media URL or local path:'
_Pause_ = 'Pause'
_Play_  = 'Play'
_Prev_  = 'Previous'
_Stop_  = 'Stop'

# GUI definition & setup
sg.theme('DarkBlue')

def Bn(name):  # a PySimpleGUI "User Defined Element" (see docs)
    return sg.Button(name, size=(8, 1), pad=(1, 1))

layout = [[sg.Input(default_text=_Path_, size=(40, 1), key='-VIDEO_PATH-'), sg.Button(_Load_)],
          [sg.Frame('', [], size=(300, 170), key='-VID_OUT-')],  # was [sg.Image('', ...)],
          [Bn(_Prev_), Bn(_Play_), Bn(_Next_), Bn(_Pause_), Bn(_Stop_)],
          [sg.Text('Load media to start', key='-MESSAGE_AREA-')]]

window = sg.Window('PySimpleGUI VLC Player', layout, element_justification='center', finalize=True, resizable=True)

window['-VID_OUT-'].expand(True, True)  # type: sg.Element

# Media Player Setup
inst = vlc.Instance()
list_player = inst.media_list_player_new()
media_list = inst.media_list_new([])
list_player.set_media_list(media_list)
player = list_player.get_media_player()
# tell VLC where to render the video(s)
tk_id = window['-VID_OUT-'].Widget.winfo_id()
libtk = ''
if sg.running_linux():
    player.set_xwindow(tk_id)
elif sg.running_windows():
    player.set_hwnd(tk_id)
elif sg.running_mac():
    try:
        from tkvlc import _GetNSView, libtk
        ns = _GetNSView(tk_id)
    except ImportError:
        ns = None
        libtk = 'none, install tkvlc.py from <https://GitHub.com/oaubert/python-vlc> examples'
    if ns:  # drawable NSview
        player.set_nsobject(ns)
    else:  # no video, only audio
        player.set_xwindow(tk_id)
else:  # running trinket, etc.
    player.set_hwnd(tk_id)  # TBD

if __name__ == '__main__':  # MCCABE 20

    if len(sys.argv) > 1:
        if sys.argv[1].lower() in ('-v', '--version'):
            # show all versions, this vlc.py, libvlc, etc. (sample output on macOS):
            # ...
            # % python3 ./psgvlc.py -v
            # psgvlc.py: 22.11.06
            # tkinter: 8.6
            # libTk: /Library/Frameworks/Python.framework/Versions/3.11/lib/libtk8.6.dylib
            # vlc.py: 3.0.12119 (Mon May 31 18:25:17 2021 3.0.12)
            # libVLC: 3.0.16 Vetinari (0x3001000)
            # plugins: /Applications/VLC.app/Contents/MacOS/plugins
            # Python: 3.11.0 (64bit) macOS 13.0 arm64
            for t in ((sys.argv[0], __version__), (sg.tk.__name__, sg.tk.TkVersion), ('libTk', libtk)):
                print('{}: {}'.format(*t))
            try:
                vlc.print_version()
                vlc.print_python()
            except AttributeError:
                pass
            sys.exit(0)

        if sys.argv[1]:
            media_list.add_media(sys.argv[1])
            list_player.set_media_list(media_list)

    # The Event Loop
    while True:
        # run with a timeout so that current location can be updated
        event, values = window.read(timeout=1000)

        if event == sg.WIN_CLOSED:
            break

        if event == _Pause_:
            list_player.pause()
        elif event == _Stop_:
            list_player.stop()
        elif event == _Next_:
            list_player.next()
            list_player.play()
        elif event == _Prev_:
            list_player.previous()  # first call causes current video to start over
            list_player.previous()  # second call moves back 1 video from current
            list_player.play()
        elif event == _Play_:
            list_player.play()
        elif event == _Load_:
            path = values['-VIDEO_PATH-']
            if path and _Path_ not in path:
                media_list.add_media(path)
                list_player.set_media_list(media_list)
                window['-VIDEO_PATH-'].update(_Path_)  # only add a legit submit

        # update elapsed time if a video loaded and playing
        if player.is_playing():
            text = '{:02d}:{:02d}'.format(*divmod(player.get_time()   // 1000, 60)) + ' / ' + \
                   '{:02d}:{:02d}'.format(*divmod(player.get_length() // 1000, 60))
            if sg.running_mac():
                print('{}: {}'.format(sys.argv[0], text))

        elif not media_list.count():
            text = 'Load media to start'
        else:
            text = 'Ready to play media'
        window['-MESSAGE_AREA-'].update(text)

    window.close()



#! /usr/bin/env python3

#
# GlSurface example code for VLC Python bindings
# Copyright (C) 2020 Daniël van Adrichem <daniel5gh@spiet.nl>

#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#

"""VLC GlSurface example
"""

import time
import ctypes
from threading import Lock

import numpy as np
from OpenGL.GL import (GL_TEXTURE_2D, glTexSubImage2D, glTexImage2D,
                       GL_BGR, GL_RGB,
                       GL_UNSIGNED_BYTE)
import vlc

class Surface(object):
    """A lockable image buffer
    """
    def __init__(self, w, h):
        self._width = w
        self._height = h

        # size in bytes when RV32 *4 or RV24 * 3
        self._row_size = self._width * 3
        self._buf_size = self._height * self._row_size
        # allocate buffer
        self._buf1 = np.zeros(self._buf_size, dtype=np.ubyte)
        # get pointer to buffer
        self._buf_p = self._buf1.ctypes.data_as(ctypes.c_void_p)
        self._lock = Lock()

    def update_gl(self):
        # with self._lock:
        glTexSubImage2D(GL_TEXTURE_2D,
                        0, 0, 0,
                        self._width,
                        self._height,
                        GL_BGR,
                        GL_UNSIGNED_BYTE,
                        self._buf1)

    def create_texture_gl(self):
        glTexImage2D(GL_TEXTURE_2D,
                     0,
                     GL_RGB,
                     self._width,  # width
                     self._height,  # height
                     0,
                     GL_BGR,
                     GL_UNSIGNED_BYTE,
                     None)

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def row_size(self):
        return self._row_size

    @property
    def buf(self):
        return self._buf1

    @property
    def buf_pointer(self):
        return self._buf_p

    def lock(self):
        self._lock.acquire()

    def unlock(self):
        self._lock.release()

    def __enter__(self, *args):
        return self._lock.__enter__(*args)

    def __exit__(self, *args):
        return self._lock.__exit__(*args)

    def get_libvlc_lock_callback(self):
        @vlc.VideoLockCb
        def _cb(opaque, planes):
            self._lock.acquire()
            planes[0] = self._buf_p

        return _cb

    def get_libvlc_unlock_callback(self):
        @vlc.VideoUnlockCb
        def _cb(opaque, picta, planes):
            self._lock.release()

        return _cb

if __name__ == '__main__':
    import sys
    player = vlc.MediaPlayer(sys.argv[1])
    # play and stop so video_get_size gets a correct value
    # setting all callbacks to None prevents a window being created on play
    player.video_set_callbacks(None, None, None, None)
    # play and stop so video_get_size gets a correct value
    player.play()
    time.sleep(1)
    player.stop()
    w, h = player.video_get_size()
    surface = Surface(w, h)
    # need to keep a reference to the CFUNCTYPEs or else it will get GCed
    _lock_cb = surface.get_libvlc_lock_callback()
    _unlock_cb = surface.get_libvlc_unlock_callback()
    player.video_set_callbacks(_lock_cb, _unlock_cb, None, None)
    player.video_set_format(
        "RV24",
        surface.width,
        surface.height,
        surface.row_size)
    # this starts populating the surface's buf with pixels, from another thread
    player.play()
    # in main thread, where gl context is current:
    # FIXME: add some code to embed the surface + a mainloop
    # v.surface.update_gl()



#! /usr/bin/env python3

#
# Qt example for VLC Python bindings
# Copyright (C) 2009-2010 the VideoLAN team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#

import sys
import os.path
import vlc
from PyQt4 import QtGui, QtCore

try:
    unicode        # Python 2
except NameError:
    unicode = str  # Python 3


class Player(QtGui.QMainWindow):
    """A simple Media Player using VLC and Qt
    """
    def __init__(self, master=None):
        QtGui.QMainWindow.__init__(self, master)
        self.setWindowTitle("Media Player")

        # creating a basic vlc instance
        self.instance = vlc.Instance()
        # creating an empty vlc media player
        self.mediaplayer = self.instance.media_player_new()

        self.createUI()
        self.isPaused = False

    def createUI(self):
        """Set up the user interface, signals & slots
        """
        self.widget = QtGui.QWidget(self)
        self.setCentralWidget(self.widget)

        # In this widget, the video will be drawn
        if sys.platform == "darwin": # for MacOS
            self.videoframe = QtGui.QMacCocoaViewContainer(0)
        else:
            self.videoframe = QtGui.QFrame()
        self.palette = self.videoframe.palette()
        self.palette.setColor (QtGui.QPalette.Window,
                               QtGui.QColor(0,0,0))
        self.videoframe.setPalette(self.palette)
        self.videoframe.setAutoFillBackground(True)

        self.positionslider = QtGui.QSlider(QtCore.Qt.Horizontal, self)
        self.positionslider.setToolTip("Position")
        self.positionslider.setMaximum(1000)
        self.connect(self.positionslider,
                     QtCore.SIGNAL("sliderMoved(int)"), self.setPosition)

        self.hbuttonbox = QtGui.QHBoxLayout()
        self.playbutton = QtGui.QPushButton("Play")
        self.hbuttonbox.addWidget(self.playbutton)
        self.connect(self.playbutton, QtCore.SIGNAL("clicked()"),
                     self.PlayPause)

        self.stopbutton = QtGui.QPushButton("Stop")
        self.hbuttonbox.addWidget(self.stopbutton)
        self.connect(self.stopbutton, QtCore.SIGNAL("clicked()"),
                     self.Stop)

        self.hbuttonbox.addStretch(1)
        self.volumeslider = QtGui.QSlider(QtCore.Qt.Horizontal, self)
        self.volumeslider.setMaximum(100)
        self.volumeslider.setValue(self.mediaplayer.audio_get_volume())
        self.volumeslider.setToolTip("Volume")
        self.hbuttonbox.addWidget(self.volumeslider)
        self.connect(self.volumeslider,
                     QtCore.SIGNAL("valueChanged(int)"),
                     self.setVolume)

        self.vboxlayout = QtGui.QVBoxLayout()
        self.vboxlayout.addWidget(self.videoframe)
        self.vboxlayout.addWidget(self.positionslider)
        self.vboxlayout.addLayout(self.hbuttonbox)

        self.widget.setLayout(self.vboxlayout)

        open = QtGui.QAction("&Open", self)
        self.connect(open, QtCore.SIGNAL("triggered()"), self.OpenFile)
        exit = QtGui.QAction("&Exit", self)
        self.connect(exit, QtCore.SIGNAL("triggered()"), sys.exit)
        menubar = self.menuBar()
        filemenu = menubar.addMenu("&File")
        filemenu.addAction(open)
        filemenu.addSeparator()
        filemenu.addAction(exit)

        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(200)
        self.connect(self.timer, QtCore.SIGNAL("timeout()"),
                     self.updateUI)

    def PlayPause(self):
        """Toggle play/pause status
        """
        if self.mediaplayer.is_playing():
            self.mediaplayer.pause()
            self.playbutton.setText("Play")
            self.isPaused = True
        else:
            if self.mediaplayer.play() == -1:
                self.OpenFile()
                return
            self.mediaplayer.play()
            self.playbutton.setText("Pause")
            self.timer.start()
            self.isPaused = False

    def Stop(self):
        """Stop player
        """
        self.mediaplayer.stop()
        self.playbutton.setText("Play")

    def OpenFile(self, filename=None):
        """Open a media file in a MediaPlayer
        """
        if filename is None:
            filename = QtGui.QFileDialog.getOpenFileName(self, "Open File", os.path.expanduser('~'))
        if not filename:
            return

        # create the media
        if sys.version < '3':
            filename = unicode(filename)
        self.media = self.instance.media_new(filename)
        # put the media in the media player
        self.mediaplayer.set_media(self.media)

        # parse the metadata of the file
        self.media.parse()
        # set the title of the track as window title
        self.setWindowTitle(self.media.get_meta(0))

        # the media player has to be 'connected' to the QFrame
        # (otherwise a video would be displayed in it's own window)
        # this is platform specific!
        # you have to give the id of the QFrame (or similar object) to
        # vlc, different platforms have different functions for this
        if sys.platform.startswith('linux'): # for Linux using the X Server
            self.mediaplayer.set_xwindow(self.videoframe.winId())
        elif sys.platform == "win32": # for Windows
            self.mediaplayer.set_hwnd(self.videoframe.winId())
        elif sys.platform == "darwin": # for MacOS
            self.mediaplayer.set_nsobject(self.videoframe.winId())
        self.PlayPause()

    def setVolume(self, Volume):
        """Set the volume
        """
        self.mediaplayer.audio_set_volume(Volume)

    def setPosition(self, position):
        """Set the position
        """
        # setting the position to where the slider was dragged
        self.mediaplayer.set_position(position / 1000.0)
        # the vlc MediaPlayer needs a float value between 0 and 1, Qt
        # uses integer variables, so you need a factor; the higher the
        # factor, the more precise are the results
        # (1000 should be enough)

    def updateUI(self):
        """updates the user interface"""
        # setting the slider to the desired position
        self.positionslider.setValue(self.mediaplayer.get_position() * 1000)

        if not self.mediaplayer.is_playing():
            # no need to call this function if nothing is played
            self.timer.stop()
            if not self.isPaused:
                # after the video finished, the play button stills shows
                # "Pause", not the desired behavior of a media player
                # this will fix it
                self.Stop()

if __name__ == "__main__":
    app = QtGui.QApplication(sys.argv)
    player = Player()
    player.show()
    player.resize(640, 480)
    if sys.argv[1:]:
        player.OpenFile(sys.argv[1])
    sys.exit(app.exec_())



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# tkinter example for VLC Python bindings
# Copyright (C) 2015 the VideoLAN team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
'''A simple example for VLC python bindings using tkinter.

Author: Patrick Fay
Date: 23-09-2015
'''

# Tested with VLC 3.0.16, 3.0.12, 3.0.11, 3.0.10, 3.0.8 and 3.0.6 with
# the compatible vlc.py Python-VLC binding, Python 3.11.0, 3.10.0, 3.9.0
# and 3.7.4 and with tkinter/Tk 8.6.9 on macOS 13.0.1 (amd64 M1), 11.6.1
# (10.16 amd64 M1), 11.0.1 (10.16 x86-64) and 10.13.6 and with VLC 3.0.18,
# Python 3.11.0 and tkinter/Tk 8.6.9 on Windows 10, all in 64-bit only.
__version__ = '22.12.28'  # mrJean1 at Gmail

import sys
try:  # Python 3.4+ only
    import tkinter as Tk
    from tkinter import TclError, ttk  # PYCHOK ttk = Tk.ttk
    from tkinter.filedialog import askopenfilename
    from tkinter.messagebox import showerror
except ImportError:
    sys.exit('%s requires Python 3.4 or later' % (sys.argv[0],))
    # import Tkinter as Tk; ttk = Tk
import os
import time
import vlc

_isMacOS   = sys.platform.startswith('darwin')
_isLinux   = sys.platform.startswith('linux')
_isWindows = sys.platform.startswith('win')

_ANCHORED    = 'Anchored'
_BANNER_H    =  32 if _isMacOS else 64
_BUTTONS     = 'Buttons'
_DISABLED    = (      Tk.DISABLED,)
_ENABLED     = ('!' + Tk.DISABLED,)
_FULL_OFF    = 'Full Off'
_FULL_SCREEN = 'Full Screen'
# see _Tk_Menu.add_item and .bind_shortcut below
# <https://www.Tcl.Tk/man/tcl8.6/TkCmd/keysyms.html>
_KEY_SYMBOL  = {'~': 'asciitilde',  '`': 'grave',
                '!': 'exclam',      '@': 'at',
                '#': 'numbersign',  '$': 'dollar',
                '%': 'percent',     '^': 'asciicirum',
                '&': 'ampersand',   '*': 'asterisk',
                '(': 'parenleft',   ')': 'parenright',
                '_': 'underscore',  '-': 'minus',
                '+': 'plus',        '=': 'equal',
                '{': 'braceleft',   '}': 'braceright',
                '[': 'bracketleft', ']': 'bracketright',
                '|': 'bar',        '\\': 'backslash',
                ':': 'colon',       ';': 'semicolon',
                '"': 'quotedbl',    "'": 'apostrophe',
                '<': 'less',        '>': 'greater',
                ',': 'comma',       '.': 'period',
                '?': 'question',    '/': 'slash',
                ' ': 'space',      '\b': 'BackSpace',  # S!
               '\n': 'KP_Enter',   '\r': 'Return',
               '\f': 'Next',       '\v': 'Prior',
               '\t': 'Tab'}  # '\a': 'space'?
# see definition of specialAccelerators in <https://
# GitHub.com/tcltk/tk/blob/main/macosx/tkMacOSXMenu.c>
_MAC_ACCEL   = {' ': 'Space',    '\b': 'Backspace',  # s!
               '\n': 'Enter',    '\r': 'Return',
               '\f': 'PageDown', '\v': 'PageUp',
               '\t': 'Tab',  # 'BackTab', 'Eject'?
            'Prior': 'PageUp', 'Next': 'PageDown'}

_MIN_W       =  420
_MOD_ALT     =  1 << 17  # alt key down?
_MOD_CMD     =  1 << 3   # command key down
_MOD_CTRL    =  1 << 2   # ctrl key down
_MOD_LOCK    =  1 << 1   # caps lock down
_MOD_SHIFT   =  1 << 0   # shift key down
_OPACITY     = 'Opacity %s%%'
_TAB_X       =  32
_T_CONFIGURE =  Tk.EventType.Configure
_T_KEY       =  Tk.EventType.Key  # KeyPress
_TICK_MS     =  100  # millisecs per time tick
_Tk_Canvas   =  Tk.Canvas
_Tk_Frame    =  ttk.Frame
_Tk_Toplevel =  Tk.Toplevel
_Tk_Version  =  Tk.TkVersion
_UN_ANCHORED = 'Un-' + _ANCHORED
_VOLUME      = 'Volume'

_TKVLC_LIBTK_PATH = 'TKVLC_LIBTK_PATH'

if _isMacOS:  # MCCABE 14
    from ctypes import cdll, c_void_p
    from ctypes.util import find_library as _find

    # libtk = cdll.LoadLibrary(ctypes.util.find_library('tk'))
    # returns (None or) the tk library /usr/lib/libtk.dylib
    # from macOS, but we need the tkX.Y library bundled with
    # Python 3+ or one matching the version of tkinter

    # Homebrew-built Python, Tcl/Tk, etc. are installed in
    # different places, usually something like (/usr/- or)
    # /opt/local/Cellar/tcl-tk/8.6.11_1/lib/libtk8.6.dylib,
    # found by command line `find /opt -name libtk8.6.dylib`

    def _find_lib(name, *paths):
        assert os.path.sep == '/'
        # 1. built into Python
        for p in (getattr(sys, 'base_prefix', ''), sys.prefix):
            if p:
                yield p + '/lib/' + name
        # 2. from ctypes.find_library, env variable
        for p in paths:
            if p:  # is not None
                p = os.path.expanduser(p)
                yield p
                if not p.endswith(name):
                    yield p + '/' + name
        # 3. the Homebrew basement
        from glob import iglob
        for t in ('/opt', '/usr'):
            t += '/local/Cellar/tcl-tk/*/lib/' + name
            for p in iglob(t):
                yield p

    try:
        env = os.environ.get(_TKVLC_LIBTK_PATH, '')
        lib = 'libtk%s.dylib' % (_Tk_Version,)
        for libtk in _find_lib(lib, _find(lib), *env.split(os.pathsep)):
            if libtk and lib in libtk and os.access(libtk, os.F_OK):
                break
        else:  # not found anywhere
            if env:  # bad env?
                t = 'no %s in %%s=%r' % (lib, env)
            else:  # env not set, suggest
                t = 'no %s found, use %%s to set a path' % (lib,)
            raise NameError(t % (_TKVLC_LIBTK_PATH,))

        lib = cdll.LoadLibrary(libtk)
        # _GetNSView = lib.TkMacOSXDrawableView is the
        # proper function to call, but that is non-public
        # (in Tk source file macosx/TkMacOSXSubwindows.c)
        # Fortunately, lib.TkMacOSXGetRootControl calls
        # lib.TkMacOSXDrawableView and returns the NSView
        _GetNSView = lib.TkMacOSXGetRootControl
        # C signature: void *_GetNSView(void *drawable) to get
        # the Cocoa/Obj-C NSWindow.contentView attribute, the
        # drawable NSView object of the (drawable) NSWindow
        _GetNSView.restype  =  c_void_p
        _GetNSView.argtypes = (c_void_p,)

    except (NameError, OSError) as x:  # lib, image or symbol not found
        libtk = str(x)  # imported by examples/psgvlc.py

        def _GetNSView(unused):  # imported by examples/psgvlc.py
            return None

    del cdll, c_void_p, env, _find, lib
    Cmd_ = 'Command+'  # bind key modifier, aka Meta_L
    # With Python 3.9+ on macOS (only!), accelerator keys specified
    # with the Shift modifier invoke the callback (command) twice,
    # once without and once with a Key (or KeyPress) event: hold the
    # former as a pseudo Key event possibly absorbed by the actual
    # Key event about 1 millisec later.  With Python 3.8- on macOS,
    # Shift accelerator keys do not work at all: do not define any
    # Shift accelerator keys in that case
    _3_9 = sys.version_info[:2] >= (3, 9)

else:  # Windows OK, untested on *nix, Xwindows
    libtk = 'N/A'
    Cmd_  = 'Ctrl+'  # bind key modifier: Control!
    _3_9  =  True


def _fullscreen(panel, *full):
    # get/set a panel full-screen or -off
    f = panel.attributes('-fullscreen')  # or .wm_attributes
    if full:
        panel.attributes('-fullscreen', bool(full[0]))
        panel.update_idletasks()
    return f


def _geometry(panel, g_w, *h_x_y):
    # set a panel geometry to C{g} or C{w, h, x, y}.
    if h_x_y:
        t = '+'.join(map(str, h_x_y))
        g = 'x'.join((str(g_w), t))
    else:
        g = g_w
    panel.geometry(g)  # update geometry, then ...
    g, *t = _geometry5(panel)  # ... get actual ...
    panel._g = g  # ... as a C{str} and 4 C{int}s
    # == panel.winfo_width(), _height(), _x(), _y()
    panel._whxy = tuple(map(int, t))
    return g


def _geometry1(panel):
    # get a panel geometry as C{str}
    panel.update_idletasks()
    return panel.geometry()


def _geometry5(panel):
    # get a panel geometry as 5-tuple of C{str}s
    g = _geometry1(panel)  # '+-x' means absolute -x
    z, x, y = g.split('+')
    w, h = z.split('x')
    return g, w, h, x, y


def _hms(tensecs, secs=''):
    # format a time (in 1/10-secs) as h:mm:ss.s
    s = tensecs * 0.1
    if s < 60:
        t = '%3.1f%s' % (s, secs)
    else:
        m, s = divmod(s, 60)
        if m < 60:
            t = '%d:%04.1f' % (int(m), s)
        else:
            h, m = divmod(m, 60)
            t = '%d:%02d:%04.1f' % (int(h), int(m), s)
    return t


def _underline2(c, label='', underline=-1, **cfg):
    # update cfg with C{underline=index} or remove C{underline=.}
    u = label.find(c) if c and label else underline
    if u >= 0:
        cfg.update(underline=u)
    else:  # no underlining
        c = ''
    cfg.update(label=label)
    return c, cfg


class _Tk_Button(ttk.Button):
    '''A C{_Tk_Button} with a label, inlieu of text.
    '''
    def __init__(self, frame, **kwds):
        cfg = self._cfg(**kwds)
        ttk.Button.__init__(self, frame, **cfg)

    def _cfg(self, label=None, **kwds):
        if label is None:
            cfg = kwds
        else:
            cfg = dict(text=label)
            cfg.update(kwds)
        return cfg

    def config(self, **kwds):
        cfg = self._cfg(**kwds)
        ttk.Button.config(self, **cfg)

    def disabled(self, *disable):
        '''Dis-/enable this button.
        '''
        # <https://TkDocs.com/tutorial/widgets.html>
        p = self.instate(_DISABLED)
        if disable:
            self.state(_DISABLED if disable[0] else _ENABLED)
        return bool(p)


class _Tk_Item(object):
    '''A re-configurable C{_Tk_Menu} item.
    '''
    def __init__(self, menu, label='', key='', under='', **kwds):
        '''New menu item.
        '''
        self.menu = menu
        self.idx  = menu.index(label)
        self.key  = key  # <...>

        self._cfg_d = dict(label=label, **kwds)
        self._dis_d = False
        self._under = under  # lower case

    def config(self, **kwds):
        '''Reconfigure this menu item.
        '''
        cfg = self._cfg_d.copy()
        cfg.update(kwds)
        if self._under:  # update underlining
            _, cfg = _underline2(self._under, **cfg)
        self.menu.entryconfig(self.idx, **cfg)

    def disabled(self, *disable):
        '''Dis-/enable this menu item.
        '''
        # <https://TkDocs.com/tutorial/menus.html>
        p = self._dis_d
        if disable:
            self._dis_d = d = bool(disable[0])
            self.config(state=Tk.DISABLED if d else Tk.NORMAL)
        return p


class _Tk_Menu(Tk.Menu):
    '''C{Tk.Menu} extended with an C{.add_shortcut} method.

       Note, make C{Command-key} shortcuts on macOS work like
       C{Control-key} shotcuts on X-/Windows using a *single*
       character shortcut.

       Other modifiers like Shift- and Option- passed thru,
       unmodified.
    '''
    _shortcuts_entries = None  # {}, see .bind_shortcuts_to
    _shortcuts_widgets = ()

    def __init__(self, master=None, **kwds):
        # remove dashed line from X-/Windows tearoff menus
        # like idlelib.editor.EditorWindow.createmenubar
        # or use root.option_add('*tearOff', False)  Off?
        # as per <https://TkDocs.com/tutorial/menus.html>
        Tk.Menu.__init__(self, master, tearoff=False, **kwds)

    def add_item(self, label='', command=None, key='', **kwds):
        '''C{Tk.menu.add_command} extended with shortcut key
           accelerator, underline and binding and returning
           a C{_Tk_Item} instance instead of an C{item} index.

           If needed use modifiers like Shift- and Alt_ or Option-
           before the *single* shortcut key character.  Do NOT
           include the Command- or Control- modifier, instead use
           the platform-specific Cmd_, like Cmd_ + key.  Also,
           do NOT enclose the key in <...> brackets since those
           are handled here as needed for the shortcut binding.
        '''
        assert callable(command), 'command=%r' % (command,)
        return self._Item(Tk.Menu.add_command, key, label,
                                  command=command, **kwds)

    def add_menu(self, label='', menu=None, key='', **kwds):  # untested
        '''C{Tk.menu.add_cascade} extended with shortcut key
           accelerator, underline and binding and returning
           a C{_Tk_Item} instance instead of an C{item} index.
        '''
        assert isinstance(menu, _Tk_Menu), 'menu=%r' % (menu,)
        return self._Item(Tk.Menu.add_cascade, key, label,
                                  menu=menu, **kwds)

    def bind_shortcut(self, key='', command=None, label='', **unused):
        '''Bind shortcut key "<modifier-...-name>".
        '''
        # C{Accelerator} modifiers on macOS are Command-,
        # Ctrl-, Option- and Shift-, but for .bind[_all] use
        # <Command-..>, <Ctrl-..>, <Option_..> and <Shift-..>
        # with a shortcut key name or character (replaced
        # with its _KEY_SYMBOL if non-alphanumeric)
        # <https://www.Tcl.Tk/man/tcl8.6/TkCmd/bind.htm#M6>
        # <https://www.Tcl.Tk/man/tcl8.6/TkCmd/keysyms.html>
        if key and callable(command) and self._shortcuts_widgets:
            for w in self._shortcuts_widgets:
                w.bind(key, command)
            if label:  # remember the key in this menu
                idx = self.index(label)
                self._shortcuts_entries[idx] = key
        # The Tk modifier for macOS' Command key is called Meta
        # with Meta_L and Meta_R for the left and right keyboard
        # keys.  Similarly for macOS' Option key, the modifier
        # name Alt with Alt_L and Alt_R.  Previously, there were
        # only the Meta_L and Alt_L keys/modifiers.  See also
        # <https://StackOverflow.com/questions/6378556/multiple-
        # key-event-bindings-in-tkinter-control-e-command-apple-e-etc>

    def bind_shortcuts_to(self, *widgets):
        '''Set widget(s) to bind shortcut keys to, usually the
           root and/or Toplevel widgets.
        '''
        self._shortcuts_entries = {}
        self._shortcuts_widgets = widgets

    def entryconfig(self, idx, command=None, **kwds):  # PYCHOK signature
        '''Update a menu item and the shortcut key binding
           if the menu item command is being changed.

           Note, C{idx} is the item's index in the menu,
           see C{_Tk_Item} above.
        '''
        if command is None:  # XXX postcommand for sub-menu
            Tk.Menu.entryconfig(self, idx, **kwds)
        elif callable(command):  # adjust the shortcut key binding
            Tk.Menu.entryconfig(self, idx, command=command, **kwds)
            key = self._shortcuts_entries.get(idx, None)
            if key is not None:
                for w in self._shortcuts_widgets:
                    w.bind(key, command)

    def _Item(self, add_, key, label, **kwds):
        # Add and bind a menu item or sub~menu with an
        # optional accelerator key (not <..> enclosed)
        # or underline letter (preceded by underscore),
        # see <https://TkDocs.com/tutorial/menus.html>.
        cfg = dict(label=label)
        if key:  # Use '+' sign, like key = "Ctrl+Shift+X"
            if key.startswith('<') and key.endswith('>'):
                c = ''  # pass as-is, e.g. <<virtual event>>
            else:
                c = '+'  # split into modifiers and char
                if key.endswith(c):
                    m = key.rstrip(c).split(c)
                else:
                    m = key.split(c)
                    c = m.pop()
                for k in ('Key', 'KeyPress', 'KeyRelease'):
                    while k in m:
                        m.remove(k)
                # adjust accelerator key for specials like KP_1,
                # PageDown and PageUp (on macOS, see function
                # ParseAccelerator in <https://GitHub.com/tcltk/tk/
                # blob/main/macosx/tkMacOSXMenu.c> and definition
                # of specialAccelerators in <https://GitHub.com/
                # tcltk/tk/blob/main/macosx/tkMacOSXMenu.c>)
                a = _MAC_ACCEL.get(c, c) if _isMacOS else c
                if a.upper().startswith('KP_'):
                    a = a[3:]
                # accelerator strings are only used for display
                # ('+' or '-' OK, ' ' space isn't for macOS)
                cfg.update(accelerator='+'.join(m + [a]))
                # replace key with Tk keysymb, allow F1 thru F35
                # (F19 on macOS) and because shortcut keys are
                # case-sensitive, use lower-case unless specified
                # as an upper-case letter with Shift+ modifier
                s = _KEY_SYMBOL.get(c, c)
                if len(s) == 1 and s.isupper() \
                               and 'Shift' not in m:
                    s = s.lower()
                # default to single key down case
                if len(c) == 1:
                    m.append('Key')  # == KeyPress
                # replace Ctrl modifier with Tk Control
                while 'Ctrl' in m:
                    m[m.index('Ctrl')] = 'Control'
                # <enclosed> for .bind_shortcut/.bind
                key = '<' + '-'.join(m + [s]) + '>'
                if _isMacOS or len(c) != 1 or not c.isalnum():
                    c = ''  # no underlining
                else:  # only Windows?
                    c, cfg = _underline2(c, **cfg)

        else:  # like idlelib, underline char after ...
            c, u = '', label.find('_')  # ... underscore
            if u >= 0:  # ... and remove underscore
                label = label[:u] + label[u+1:]
                cfg.update(label=label)
                if u < len(label) and not _isMacOS:
                    # c = label[u]
                    cfg.update(underline=u)

        if kwds:  # may still override accelerator ...
            cfg.update(kwds)  # ... and underline
        add_(self, **cfg)  # first _add then ...
        self.bind_shortcut(key, **cfg)  # ... bind
        return _Tk_Item(self, key=key, under=c, **cfg)


class _Tk_Slider(Tk.Scale):
    '''Scale with some add'l attributres
    '''
    _var = None

    def __init__(self, frame, to=1, **kwds):
        if isinstance(to, int):
            f, v = 0, Tk.IntVar()
        else:
            f, v = 0.0, Tk.DoubleVar()
        cfg = dict(from_=f, to=to,
                   orient=Tk.HORIZONTAL,
                   showvalue=0,
                   variable=v)
        cfg.update(kwds)
        Tk.Scale.__init__(self, frame, **cfg)
        self._var = v

    def set(self, value):
        # doesn't move the slider
        self._var.set(value)
        Tk.Scale.set(self, value)


class Player(_Tk_Frame):
    '''The main window handling with events, etc.
    '''
    _anchored  =  True  # under the video panel
    _BUTTON_H  = _BANNER_H
    _debugs    =  0
    _isFull    = ''  # or geometry
    _length    =  0  # length time ticks
    _lengthstr = ''  # length h:m:s
    _muted     =  False
    _opacity   =  90 if _isMacOS else 100  # percent
    _opaque    =  False
    _rate      =  0.0
    _ratestr   = ''
    _scaleX    =  1
    _scaleXstr = ''
    _sliding   =  False
    _snapshots =  0
    _stopped   =  None
    _title     = 'tkVLCplayer'
    _volume    =  50  # percent

    def __init__(self, parent, title='', video='', debug=False):  # PYCHOK called!
        _Tk_Frame.__init__(self, parent)

        self.debug  = bool(debug)
        self.parent = parent  # == root
        self.video  = os.path.expanduser(video)
        if title:
            self._title = str(title)
        parent.title(self._title)
#       parent.iconname(self._title)

        # set up tickers to avoid None error
        def _pass():
            pass

        self._tick_a = self.after(1, _pass)
        self._tick_c = self.after(2, _pass)
        self._tick_r = self.after(3, _pass)
        self._tick_s = self.after(4, _pass)  # .after_idle
        self._tick_t = self.after(5, _pass)
        self._tick_z = self.after(6, _pass)

        # panels to play videos and hold buttons, sliders,
        # created *before* the File menu to be able to bind
        # the shortcuts keys to both windows/panels.
        self.videoPanel = v = self._VideoPanel()
        self._bind_events(v)  # or parent
        self.buttonsPanel = b = self._ButtonsPanel()
        self._bind_events(b)

        mb = _Tk_Menu(self.parent)  # menu bar
        parent.config(menu=mb)
#       self.menuBar = mb

        # macOS shortcuts <https://Support.Apple.com/en-us/HT201236>
        m = _Tk_Menu(mb)  # Video menu, shortcuts to both panels
        m.bind_shortcuts_to(v, b)
        m.add_item('Open...', self.OnOpen, key=Cmd_ + 'O')
        m.add_separator()
        self.playItem = m.add_item('Play', self.OnPlay, key=Cmd_ + 'P')  # Play/Pause
        m.add_item('Stop', self.OnStop, key=Cmd_ + '\b')  # BackSpace
        m.add_separator()
        m.add_item('Zoom In',  self.OnZoomIn,  key=(Cmd_ + 'Shift++') if _3_9 else '')
        m.add_item('Zoom Out', self.OnZoomOut, key=(Cmd_ + '-')       if _3_9 else '')
        m.add_separator()
        m.add_item('Faster', self.OnFaster, key=(Cmd_ + 'Shift+>') if _3_9 else '')
        m.add_item('Slower', self.OnSlower, key=(Cmd_ + 'Shift+<') if _3_9 else '')
        m.add_separator()
        m.add_item('Normal 1X', self.OnNormal, key=Cmd_ + '=')
        m.add_separator()
        self.muteItem = m.add_item('Mute', self.OnMute, key=Cmd_ + 'M')
        m.add_separator()
        m.add_item('Snapshot', self.OnSnapshot, key=Cmd_ + 'T')
        m.add_separator()
        self.fullItem = m.add_item(_FULL_SCREEN, self.OnFull, key=Cmd_ + 'F')
        m.add_separator()
        m.add_item('Close', self.OnClose, key=Cmd_ + 'W')
        mb.add_cascade(menu=m, label='Video')
#       self.videoMenu = m

        m = _Tk_Menu(mb)  # Video menu, shortcuts to both panels
        m.bind_shortcuts_to(v, b)
        self.anchorItem = m.add_item(_UN_ANCHORED, self.OnAnchor, key=Cmd_ + 'A')
        m.add_separator()
        self.opaqueItem = m.add_item(_OPACITY % (self._opacity,), self.OnOpacity, key=Cmd_ + 'Y')
        m.add_item('Normal 100%', self.OnOpacity100)
        mb.add_cascade(menu=m, label=_BUTTONS)
#       self.buttonsMenu = m

        if _isMacOS and self.debug:  # Special macOS "windows" menu
            # <https://TkDocs.com/tutorial/menus.html> "Providing a Window Menu"
            # XXX Which (virtual) events are generated other than Configure?
            m = _Tk_Menu(mb, name='window')  # must be name='window'
            mb.add_cascade(menu=m, label='Windows')

        # VLC player
        args = ['--no-xlib'] if _isLinux else []
        self.Instance = vlc.Instance(args)
        self.player = self.Instance.media_player_new()

        b.update_idletasks()
        v.update_idletasks()
        if self.video:  # play video for a second, adjusting the panel
            self._play(self.video)
            self.after(1000, self.OnPause)
#       elif _isMacOS:  # <https://StackOverflow.com/questions/18394597/
#           # is-there-a-way-to-create-transparent-windows-with-tkinter>
#           self._stopped = True
#           self._set_opacity()
        self.OnTick()  # set up the timer

        # Keep the video panel at least as wide as the buttons panel
        # and move it down enough to put the buttons panel above it.
        self._BUTTON_H = d = b.winfo_height()
        b.minsize(width=_MIN_W, height=d)
        v.minsize(width=_MIN_W, height=0)
        _, w, h, _, y = _geometry5(v)
        y = int(y) + d + _BANNER_H
        _geometry(v, w, h, _TAB_X, y)
        self._anchorPanels()
        self._set_volume()

    def _anchorPanels(self, video=False):
        # Put the buttons panel under the video
        # or the video panel above the buttons
        if self._anchored and not self._isFull:
            self._debug(self._anchorPanels)
            v = self.videoPanel
            if _isMacOS and _fullscreen(v):
                # macOS green button full-screen?
                _fullscreen(v, False)
                self.OnFull()
            else:
                b = self.buttonsPanel
                v.update_idletasks()
                b.update_idletasks()
                h = v.winfo_height()
                d = h + _BANNER_H  # vertical delta
                if video:  # move/adjust video panel
                    w = b.winfo_width()  # same as ...
                    x = b.winfo_x()      # ... buttons
                    y = b.winfo_y() - d  # ... and above
                    g = v
                else:  # move/adjust buttons panel
                    h = b.winfo_height()  # unchanged
                    if h > self._BUTTON_H and _fullscreen(b):
                        # macOS green button full-screen?
                        _fullscreen(b, False)
                        h = self._BUTTON_H
                    w = v.winfo_width()  # unchanged
                    x = v.winfo_x()  # same as the video
                    y = v.winfo_y() + d  # below the video
                    g = b
#               _g = g._g
                _geometry(g, max(w, _MIN_W), h, x, y)
                if video:  # and g._g != _g:
                    self._set_aspect_ratio(True)

    def _bind_events(self, panel):
        # set up handlers for several events
        try:
            p  = panel
            p_ = p.protocol
        except AttributeError:
            p  = p.master  # == p.parent
            p_ = p.protocol
        if _isWindows:  # OK for macOS
            p_('WM_DELETE_WINDOW', self.OnClose)
#       Event Types <https://www.Tcl.Tk/man/tcl8.6/TkCmd/bind.html#M7>
        p.bind('<Configure>',       self.OnConfigure)  # window resize, position, etc.
        # needed on macOS to catch window close events
        p.bind('<Destroy>',         self.OnClose)      # window half-dead
#       p.bind('<Activate>',        self.OnActive)     # window activated
#       p.bind('<Deactivate>',      self.OffActive)    # window deactivated
        p.bind('<FocusIn>',         self.OnFocus)      # getting keyboard focus
#       p.bind('<FocusOut>',        self.OffFocus)     # losing keyboard focus
#       p.bind('<Return>',          self.OnReturn)     # highlighted button
        if _isMacOS:
            p.bind('<Command-.>', self.OnClose)
#           p.bind('<Command-,.'. self.OnPreferences)
#           p.bind('<KP_Enter>',  self.OnReturn)  # highlighted button
        # attrs holding the most recently set _geometry ...
        assert not hasattr(panel, '_g')
        panel._g = ''  # ... as a sC{str} and ...
        assert not hasattr(panel, '_whxy')
        panel._whxy = ()  # ... 4-tuple of C{ints}s

    def _ButtonsPanel(self):
        # create panel with buttons and sliders
        b = _Tk_Toplevel(self.parent, name='buttons')
        t = '%s - %s' % (self._title, _BUTTONS)
        b.title(t)  # '' removes the window banner
        b.resizable(True, False)

        f = _Tk_Frame(b)
        # button are too small on Windows if width is given
        p = _Tk_Button(f, label='Play', command=self.OnPlay)
        #                 width=len('Pause'), underline=0
        s = _Tk_Button(f, label='Stop', command=self.OnStop)
        m = _Tk_Button(f, label='Mute', command=self.OnMute)
        #                 width=len('Unmute'), underline=0
        q = _Tk_Slider(f, command=self.OnPercent, to=100,
                          label=_VOLUME)  # length=170
        p.pack(side=Tk.LEFT, padx=8)
        s.pack(side=Tk.LEFT)
        m.pack(side=Tk.LEFT, padx=8)
        q.pack(fill=Tk.X,    padx=4, expand=1)
        f.pack(side=Tk.BOTTOM, fill=Tk.X)

        f = _Tk_Frame(b)  # new frame?
        t = _Tk_Slider(f, command=self.OnTime, to=1000,  # millisecs
                          length=_MIN_W)  # label='Time'
        t.pack(side=Tk.BOTTOM, fill=Tk.X, expand=1)
        f.pack(side=Tk.BOTTOM, fill=Tk.X)

        # <https://www.PythonTutorial.net/tkinter/tkinter-window>
        # <https://TkDocs.com/tutorial/windows.html>
        # b.attributes('-topmost', 1)

        # self.videoPanel.update()  # needed to ...
# #     b.overrideredirect(True)  # ignore the panel
        b.update_idletasks()

        self.muteButton    = m
        self.playButton    = p
        self.timeSlider    = t
        self.percentSlider = q
        return b

    def _debug(self, where, *event, **kwds):
        # Print where an event is are handled.
        if self.debug:
            self._debugs += 1
            d = dict(anchored=self._anchored,
                       isFull=bool(self._isFull),
                      opacity=self._opacity,
                       opaque=self._opaque,
                      stopped=self._stopped,
                       volume=self._volume)
            p = self.player
            if p and p.get_media():
                d.update(playing=p.is_playing(),
                            rate=p.get_rate(),
                           scale=p.video_get_scale(),
                           scaleX=self._scaleX)
            try:  # final OnClose may throw TclError
                d.update(Buttons=_geometry1(self.buttonsPanel))
                d.update(  Video=_geometry1(self.videoPanel))
                if event:  # an event
                    event = event[0]
                    d.update(event=event)
                    w = str(event.widget)
#                   d.update(widget=type(event.widget))  # may fail
                    d.update(Widget={'.':        'Video',
                                     '.buttons': _BUTTONS}.get(w, w))
            except (AttributeError, TclError):
                pass
            d.update(kwds)
            d = ', '.join('%s=%s' % t for t in sorted(d.items()))
            print('%4s: %s %s' % (self._debugs, where.__name__, d))

    def _frontmost(self):
        # Move panels to the front ...  temporarily.
        for p in (self.videoPanel, self.buttonsPanel):
            p.attributes('-topmost', True)
            p.update_idletasks()
            p.attributes('-topmost', False)
            try:  # no Toplevel.force_focus
                p.force_focus()
            except AttributeError:
                pass

    def OnAnchor(self, *unused):
        '''Toggle anchoring of the panels.
        '''
        self._debug(self.OnAnchor)
        self._anchored = not self._anchored
        if self._anchored:
            self._anchorPanels()
            a = _UN_ANCHORED
        else:  # move the buttons panel to the top-left
            b = self.buttonsPanel
            h = b.winfo_height()  # unchanged
            _geometry(b, _MIN_W, h, _TAB_X, _BANNER_H)
            a = _ANCHORED
        self.anchorItem.config(label=a)

    def OnClose(self, *event):
        '''Closes the window(s) and quit.
        '''
        self._debug(self.OnClose, *event)
        # print('_quit: bye')
        self.after_cancel(self._tick_a)
        self.after_cancel(self._tick_c)
        self.after_cancel(self._tick_r)
        self.after_cancel(self._tick_s)
        self.after_cancel(self._tick_t)
        self.after_cancel(self._tick_z)
        v = self.videoPanel
        v.update_idletasks()
        self.quit()  # stops .mainloop

    def OnConfigure(self, event):
        '''Some widget configuration changed.
        '''
        w, T = event.widget, event.type  # int
        if T == _T_CONFIGURE and w.winfo_toplevel() is w:
            # i.e. w is videoFrame/Panel or buttonsPanel
            if w is self.videoPanel:
                a = self._set_aspect_ratio  # force=True
            elif w is self.buttonsPanel and self._anchored:
                a = self._anchorPanels  # video=True
            else:
                a = None
            # prevent endless, recursive onConfigure events due to changing
            # the buttons- and videoPanel geometry, especially on Windows
            if a and w._whxy != (event.width, event.height, event.x, event.y):
                self.after_cancel(self._tick_c)
                self._debug(self.OnConfigure, event)
                self._tick_c = self.after(250, a, True)

    def OnFaster(self, *event):
        '''Speed the video up by 25%.
        '''
        self._set_rate(1.25, *event)
        self._debug(self.OnFaster)

    def OnFocus(self, *unused):
        '''Got the keyboard focus.
        '''
        self._debug(self.OnFocus)
        self._frontmost()
#       self._set_aspect_ratio()
#       self._wiggle()

    def OnFull(self, *unused):
        '''Toggle full/off screen.
        '''
        self._debug(self.OnFull)
        # <https://www.Tcl.Tk/man/tcl8.6/TkCmd/wm.htm#M10>
        # self.after_cancel(self._tick_t)
        v = self.videoPanel
        if not _fullscreen(v):
            self._isFull = _geometry1(v)
            _fullscreen(v, True)  # or .wm_attributes
            v.bind('<Escape>', self.OnFull)
            f = _FULL_OFF
        else:
            _fullscreen(v, False)
            v.unbind('<Escape>')
            _geometry(v, self._isFull)
            self._isFull = ''
            self._anchorPanels()
            f = _FULL_SCREEN
        self.fullItem.config(label=f)

    def OnMute(self, *unused):
        '''Mute/Unmute audio.
        '''
        if self._stopped or self._opaque:
            return  # button.disabled
        self._debug(self.OnMute)
        # audio un/mute may be unreliable, see vlc.py docs.
        self._muted = m = not self._muted  # self.player.audio_get_mute()
        self.player.audio_set_mute(m)
        u = 'Unmute' if m else 'Mute'
        # i = u.index('m' if m else 'M')  # 2 if m else 0
        self.muteItem.config(label=u)
        self.muteButton.config(label=u)  # width=len(u), underline=i
        self.OnPercent()  # re-label the slider

    def OnNormal(self, *unused):
        '''Normal speed and 1X zoom.
        '''
        self._frontmost()
        self._set_rate(0.0)
        self._set_zoom(0.0)
#       self._wiggle()
        self._set_aspect_ratio(True)
        self._debug(self.OnNormal)

    def OnOpacity(self, *unused):
        '''Use the percent slider to adjust the opacity.
        '''
        self.muteButton.disabled(True)  # greyed out?
        self.muteItem.disabled(True)  # greyed out?
        self._opaque = True
        self._set_opacity()
        self._debug(self.OnOpacity)

    def OnOpacity100(self, *unused):
        '''Set the opacity to 100%.
        '''
        self._frontmost()
        self._set_opacity(100)
        self._debug(self.OnOpacity100)

    def OnOpen(self, *unused):
        '''Show the file dialog to choose a video, then play it.
        '''
        self._debug(self.OnOpen)
        self._reset()
        # XXX ... +[CATransaction synchronize] called within transaction
        v = askopenfilename(initialdir=os.path.expanduser('~'),
                            title='Choose a video',
                            filetypes=(('all files', '*.*'),
                                       ('mp4 files', '*.mp4'),
                                       ('mov files', '*.mov')))
        self._play(os.path.expanduser(v))
        self._set_aspect_ratio(True)

    def OnPause(self, *unused):
        '''Toggle between Pause and Play.
        '''
        self._debug(self.OnPause)
        p = self.player
        if p.get_media():
            self._pause_play(not p.is_playing())
#           self._wiggle()
            p.pause()  # toggles

    def OnPercent(self, *unused):
        '''Percent slider changed, adjust the opacity or volume.
        '''
        self._debug(self.OnPercent)
        s = max(0, min(100, self.percentSlider.get()))
        if self._opaque or self._stopped:
            self._set_opacity(s)
        else:
            self._set_volume(s)

    def OnPlay(self, *unused):
        '''Play video, if there's no video to play or
           playing, show a Tk.FileDialog to select one
        '''
        self._debug(self.OnPlay)
        p = self.player
        m = p.get_media()
        if not m:
            if self.video:
                self._play(self.video)
                self.video = ''
            else:
                self.OnOpen()
        elif p.play():  # == -1, play failed
            self._showError('play ' + repr(m))
        else:
            self._pause_play(True)
            if _isMacOS:
                self._wiggle()

    def OnSlower(self, *event):
        '''Slow the video down by 20%.
        '''
        self._set_rate(0.80, *event)
        self._debug(self.OnSlower)

    def OnSnapshot(self, *unused):
        '''Take a snapshot and save it (as .PNG only).
        '''
        p = self.player
        if p and p.get_media():
            self._snapshots += 1
            S = 'Snapshot%s' % (self._snapshots,)
            s = '%s-%s.PNG' % (self._title, S)  # PNG only
            if p.video_take_snapshot(0, s, 0, 0):
                self._showError('take ' + S)

    def OnStop(self, *unused):
        '''Stop the player, clear panel, etc.
        '''
        self._debug(self.OnStop)
        self._reset()

    def OnTick(self):
        '''Udate the time slider with the video time.
        '''
        p = self.player
        if p:
            s = self.timeSlider
            if self._length > 0:
                if not self._sliding:  # see .OnTime
                    t = max(0, p.get_time() // _TICK_MS)
                    if t != s.get():
                        s.set(t)
                        self._set_buttons_title(t)
            else:  # get video length in millisecs
                t = p.get_length()
                if t > 0:
                    self._length = t = max(1, t // _TICK_MS)
                    self._lengthstr = _hms(t, secs=' secs')
                    s.config(to=t)  # tickinterval=t / 5)
        # re-start the 1/4-second timer
        self._tick_t = self.after(250, self.OnTick)

    def OnTime(self, *unused):
        '''Time slider has been moved by user.
        '''
        if self.player and self._length:
            self._sliding = True  # slider moving, see .OnTick
            self.after_cancel(self._tick_s)
            t = self.timeSlider.get()
            self._tick_s = self.after_idle(self._set_time, t * _TICK_MS)
            self._set_buttons_title(t)
            self._debug(self.OnTime, tensecs=t)

    def OnZoomIn(self, *event):
        '''Zoom in by 25%.
        '''
        self._set_zoom(1.25, *event)
        self._debug(self.OnZoomIn)

    def OnZoomOut(self, *event):
        '''Zoom out by 20%.
        '''
        self._set_zoom(0.80, *event)
        self._debug(self.OnZoomOut)

    def _pause_play(self, playing):
        # re-label menu item and button, adjust callbacks
        p = 'Pause' if playing else 'Play'
        c = self.OnPlay if playing is None else self.OnPause  # PYCHOK attr
        self.playButton.config(label=p, command=c)
        self.playItem.config(label=p, command=c)
        self.muteButton.disabled(False)
        self.muteItem.disabled(False)
        self._stopped = self._opaque = False
        self._set_buttons_title()
        self._set_opacity()  # no re-label
        self._set_volume()
        self._set_aspect_ratio(True)

    def _play(self, video):
        # helper for OnOpen and OnPlay
        if os.path.isfile(video):  # Creation
            m = self.Instance.media_new(str(video))  # unicode
            p = self.player
            p.set_media(m)
            t = '%s - %s' % (self._title, os.path.basename(video))
            self.videoPanel.title(t)
#           self.buttonsPanel.title(t)

            # get the window handle for VLC to render the video
            h = self.videoCanvas.winfo_id()  # .winfo_visualid()?
            if _isWindows:
                p.set_hwnd(h)
            elif _isMacOS:
                # (1) the handle on macOS *must* be an NSView
                # (2) the video fills the entire panel, covering
                # all frames, buttons, sliders, etc. inside it
                ns = _GetNSView(h)
                if ns:
                    p.set_nsobject(ns)
                else:  # no libtk: no video, only audio
                    p.set_xwindow(h)
            else:  # *nix, Xwindows
                p.set_xwindow(h)  # fails on Windows
            self.OnPlay(None)

    def _reset(self):
        # stop playing, clear panel
        p = self.player
        if p:
            p.stop()
            self.timeSlider.set(0)
            self._pause_play(None)
            self._sliding = False
            self._stopped = True
            self.OnOpacity()

    def _set_aspect_ratio(self, force=False):
        # set the video panel aspect ratio and re-anchor
        p = self.player
        if p and not self._isFull:
            v = self.videoPanel
            g, w, h, x, y = _geometry5(v)
            if force or g != v._g:  # update
                self.after_cancel(self._tick_a)
                a, b = p.video_get_size(0)  # often (0, 0)
                if b > 0 and a > 0:
                    # adjust the video panel ...
                    if a > b:  # ... landscape height
                        h = round(float(w) * b / a)
                    else:  # ... or portrait width
                        w = round(float(h) * a / a)
                    _g = _geometry(v, w, h, x, y)
                    self._debug(self._set_aspect_ratio, a=a, b=b)
                    if self._anchored and (force or _g != g):
                        self._anchorPanels()
                # redo periodically since (1) player.video_get_size()
                # only returns non-zero width and height after playing
                # for a while and (2) avoid too frequent updates during
                # manual resizing of the video panel
                self._tick_a = self.after(500, self._set_aspect_ratio)

    def _set_buttons_title(self, *tensecs):
        # set the buttons panel title
        T, s = self._length, self._lengthstr
        if s and T:
            t =  tensecs[0] if tensecs else self.timeSlider.get()
            t = _hms(t) if t < T else s
            t = '%s - %s / %s%s%s' % (self._title, t, s, self._scaleXstr, self._ratestr)
        else:  # reset panel title
            t = '%s - %s' % (self._title, _BUTTONS)
            self._length = 0
#           self._lengthstr = ''
        self.buttonsPanel.title(t)

    def _set_opacity(self, *percent):  # 100% fully opaque
        # set and re-label the opacity, panels and menu item
        if percent:
            self._opacity = p = percent[0]
        else:
            p = self._opacity
        a = max(0.2, min(1, p * 1e-2))
        self.videoPanel.attributes('-alpha', a if self._stopped else 1)
        self.buttonsPanel.attributes('-alpha', a)
#       if _isMacOS:  # <https://TkDocs.com/tutorial/windows.html>
#           self.buttonsPanel.attributes('-transparent', a)
        s = _OPACITY % (p,)
        self.opaqueItem.config(label=s)
        if self._opaque or self._stopped:
            self._set_percent(p, label=s)

    def _set_percent(self, percent, **cfg):
        # set and re-label the slider
        self.percentSlider.config(**cfg)
        self.percentSlider.set(percent)

    def _set_rate(self, factor, *event):
        # change the video rate
        p = self.player
        self.after_cancel(self._tick_r)
        if not event:  # delay the menu event as a false key event ...
            # ... and possibly overwritten by the actual key event
            self._tick_r = self.after(3, self._set_rate, factor, False)
        elif p:
            r = p.get_rate() * factor
            if r > 0:
                r = max(0.2, min(10.0, r))
                t = ' - %d%%' % (int(r * 100),)
            else:
                r, t = 1.0, ''
            p.set_rate(r)
            self._rate = r
            self._ratestr = t
            self._set_buttons_title()

    def _set_time(self, millisecs):
        # set player to time
        p = self.player
        if p:
            p.set_time(millisecs)
        self._sliding = False  # see .OnTick

    def _set_volume(self, *volume):
        # set and re-label the volume
        if volume:
            self._volume = v = volume[0]
        else:
            v = self._volume
        m = ' (Muted)' if self._muted else ''
        V = '%s %s%%' % (_VOLUME, v)
        self._set_percent(v, label=V + m)
        p = self.player
        if p and p.is_playing() and not self._stopped:
            # .audio_set_volume returns 0 on success, -1 otherwise,
            # e.g. if the player is stopped or doesn't have media
            if p.audio_set_volume(v):  # and p.get_media():
                self._showError('set ' + V)

    def _set_zoom(self, factor, *event):
        # zoom the video in/out, see cocoavlc.py
        p = self.player
        self.after_cancel(self._tick_z)
        if not event:  # delay the menu event as a false key event ...
            # ... and possibly overwritten by the actual key event
            self._tick_z = self.after(3, self._set_zoom, factor, False)
        elif p:
            x = self._scaleX * factor
            if x > 1:
                s = x
                t = ' - %.1fX' % (x,)
            else:
                x, s, t = 1, 0.0, ''
            p.video_set_scale(s)
#           self.videoPanel.update_idletasks()
            self._scaleX = x
            self._scaleXstr = t
            self._set_buttons_title()

    def _showError(self, verb):
        '''Display a simple error dialog.
        '''
        t = 'Unable to %s' % (verb,)
        showerror(self._title, t)
        # sys.exit(t)

    def _VideoPanel(self):
        # create panel to play video
        v = _Tk_Frame(self.parent)
        c = _Tk_Canvas(v)  # takefocus=True
        c.pack(fill=Tk.BOTH, expand=1)
        v.pack(fill=Tk.BOTH, expand=1)
        v.update_idletasks()

        self.videoCanvas = c
        self.videoFrame  = v
        # root is used for updates, NOT ...
        return self.parent  # ... the frame

    def _wiggle(self, d=4):
        # wiggle the video to fill the window on macOS
        if not self._isFull:
            v = self.videoPanel
            g, w, h, x, y = _geometry5(v)
            w = int(w) + d
            # x = int(x) - d
            # h = int(h) + d
            if _geometry(v, w, h, x, y) != g:
                self.after_idle(_geometry, v, g)
        if d > 1:  # repeat a few times
            self.after(100, self._wiggle, d - 1)


def print_version(name=''):  # imported by psgvlc.py
    # show all versions, this module, tkinter, libtk, vlc.py, libvlc, etc.

    # sample output on macOS:

    # % python3 ./tkvlc.py -v
    # tkvlc.py: 22.12.28
    # tkinter: 8.6
    # libTk: /Library/Frameworks/Python.framework/Versions/3.11/lib/libtk8.6.dylib
    # is_TK: aqua, isAquaTk, isCocoaTk
    # vlc.py: 3.0.12119 (Mon May 31 18:25:17 2021 3.0.12)
    # libVLC: 3.0.16 Vetinari (0x3001000)
    # plugins: /Applications/VLC.app/Contents/MacOS/plugins
    # Python: 3.11.0 (64bit) macOS 13.0.1 arm64

    # sample output on Windows:

    # PS C: python3 .\tkvlc.py -v
    # tkvlc.py: 22.12.28
    # tkinter: 8.6
    # libTk: N/A
    # is_TK: win32
    # vlc.py: 3.0.12119 (Mon May 31 18:25:17 2021 3.0.12)
    # libVLC: 3.0.18 Vetinari (0x3001200)
    # plugins: C:\Program Files\VideoLAN\VLC
    # Python: 3.11.0 (64bit) Windows 10

    # see <https://TkDocs.com/tutorial/menus.html> or private property
    r = Tk.Tk()
    t = r.tk.call('tk', 'windowingsystem'),  # r._windowingsystem
    r.destroy()
    if _isMacOS:
        try:
            from idlelib import macosx
            m  = macosx.__dict__
            t += tuple(sorted(n for n, t in m.items() if n.startswith('is') and
                                                         n.endswith('Tk') and
                                                         callable(t) and t()))
        except ImportError:  # Python 10: no test.support ...
            pass
    t = ', '.join(t) or 'N/A'

    n = os.path.basename(name or __file__)
    for t in ((n, __version__), (Tk.__name__, _Tk_Version), ('libTk', libtk), ('is_Tk', t)):
        print('%s: %s' % t)

    try:
        vlc.print_version()
        vlc.print_python()
    except AttributeError:
        try:
            os.system(sys.executable + ' -m vlc -v')
        except OSError:
            pass


if __name__ == '__main__':  # MCCABE 13

    _argv0 = sys.argv[0]
    _debug = False
    _video = ''

    while len(sys.argv) > 1:
        arg = sys.argv.pop(1)
        if arg in ('-v', '--version'):
            print_version()
            sys.exit(0)
        elif '-debug'.startswith(arg) and len(arg) > 3:
            _debug = True
        elif arg.startswith('-'):
            print('usage: %s  [-v | --version]  [-debug]  [<video_file_name>]' % (_argv0,))
            sys.exit(1)
        elif arg:  # video file
            _video = os.path.expanduser(arg)
            if not os.path.isfile(_video):
                print('%s error, no such file: %r' % (_argv0, arg))
                sys.exit(1)

    root = Tk.Tk()  # create a Tk.App()
    player = Player(root, video=_video, debug=_debug)
    if _isWindows:  # see function _test() at the bottom of ...
        # <https://GitHub.com/python/cpython/blob/3.11/Lib/tkinter/__init__.py>
        root.iconify()
        root.update()
        root.deiconify()
        root.mainloop()  # forever
        root.destroy()  # this is necessary on Windows to avoid ...
        # ... Fatal Python Error: PyEval_RestoreThread: NULL tstate
    else:
        root.mainloop()  # forever



from distribute_setup import use_setuptools
use_setuptools()

from setuptools import setup

setup(name='python-vlc',
      version = '3.0.20123',
      author='Olivier Aubert',
      author_email='contact@olivieraubert.net',
      maintainer='Olivier Aubert',
      maintainer_email='contact@olivieraubert.net',
      url='http://wiki.videolan.org/PythonBinding',
      py_modules=['vlc'],
      keywords = [ 'vlc', 'video' ],
      license = "LGPL-2.1+",
      classifiers = [
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
          "Operating System :: MacOS :: MacOS X",
          "Operating System :: Microsoft :: Windows",
          "Operating System :: POSIX :: Linux",
          "Operating System :: POSIX :: Other",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 3",
          "Topic :: Multimedia",
          "Topic :: Multimedia :: Sound/Audio",
          "Topic :: Multimedia :: Video",
      ],
      description = "VLC bindings for python.",
      long_description = """This module provides ctypes-based bindings (see
      http://wiki.videolan.org/PythonBinding) for the native libvlc
      API (see http://wiki.videolan.org/LibVLC) of the VLC video
      player. Note that it relies on an already present install of VLC.

      It has been automatically generated from the include files of
      vlc 3.0.20, using generator 1.23.
      """)



#!/usr/bin/env python3

# Author:   A.Invernizzi (@albestro on GitHub)
# Date:     Jun 03, 2020

# MIT License <http://OpenSource.org/licenses/MIT>
#
# Copyright (C) 2020 -- A. Invernizzi, @albestro on github
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

"""
Example usage of VLC API function `libvlc_media_new_callbacks`
This function allows to create a VLC media `libvlc_media_t` specifying custom
callbacks where the user can define how to manage and read the stream of data.

The general use case for this is when you have data in memory and you want to
play it (e.g. audio stream from a web radio).

In this example, we are going to read playable data from files in a specified
folder. In case you would want to read from a file, it is not the best way to do it,
but for the sake of this example we are going to read data into memory from files.

The example tries to highlight the separation of concerns between the callbacks and
the application logic, so it would hopefully make clear how to integrate the VLC API
with existing libraries.

In particular, we have two main parts:
    - StreamProvider: which is a class that implements the logic; "scrape" a folder
    for files with a specific extensions, and provide methods that retrieves data.
    - VLC callabacks that uses a StreamProvider object
"""

import argparse
import ctypes
import os

import vlc


class StreamProviderDir(object):
    def __init__(self, rootpath, file_ext):
        self._media_files = []
        self._rootpath = rootpath
        self._file_ext = file_ext
        self._index = 0

    def open(self):
        """
        this function is responsible of opening the media.
        it could have been done in the __init__, but it is just an example

        in this case it scan the specified folder, but it could also scan a
        remote url or whatever you prefer.
        """

        print("read file list")
        for entry in os.listdir(self._rootpath):
            if os.path.splitext(entry)[1] == f".{self._file_ext}":
                self._media_files.append(os.path.join(self._rootpath, entry))
        self._media_files.sort()

        print("playlist:")
        for index, media_file in enumerate(self._media_files):
            print(f"[{index}] {media_file}")

    def release_resources(self):
        """
        In this example this function is just a placeholder,
        in a more complex example this may release resources after the usage,
        e.g. closing the socket from where we retrieved media data
        """
        print("releasing stream provider")

    def seek(self, offset):
        """
        Again, a placeholder, not useful for the example
        """
        print(f"requested seek with offset={offset}")

    def get_data(self):
        """
        It reads the current file in the list and returns the binary data
        In this example it reads from file, but it could have downloaded data from an url
        """
        print(f"reading file [{self._index}] ", end='')

        if self._index == len(self._media_files):
            print("file list is over")
            return b''

        print(f"{self._media_files[self._index]}")
        with open(self._media_files[self._index], 'rb') as stream:
            data = stream.read()

        self._index = self._index + 1

        return data


# HERE THERE ARE THE CALLBACKS USED BY THE MEDIA CREATED IN THE "MAIN"
# a callback in its simplest form is a python function decorated with the specific @vlc.CallbackDecorators.*

@vlc.CallbackDecorators.MediaOpenCb
def media_open_cb(opaque, data_pointer, size_pointer):
    print("OPEN", opaque, data_pointer, size_pointer)

    stream_provider = ctypes.cast(opaque, ctypes.POINTER(ctypes.py_object)).contents.value

    stream_provider.open()

    data_pointer.contents.value = opaque
    size_pointer.value = 1 ** 64 - 1

    return 0


@vlc.CallbackDecorators.MediaReadCb
def media_read_cb(opaque, buffer, length):
    print("READ", opaque, buffer, length)

    stream_provider = ctypes.cast(opaque, ctypes.POINTER(ctypes.py_object)).contents.value

    new_data = stream_provider.get_data()
    bytes_read = len(new_data)

    if bytes_read > 0:
        buffer_array = ctypes.cast(buffer, ctypes.POINTER(ctypes.c_char * bytes_read))
        for index, b in enumerate(new_data):
            buffer_array.contents[index] = ctypes.c_char(b)

    print(f"just read f{bytes_read}B")
    return bytes_read


@vlc.CallbackDecorators.MediaSeekCb
def media_seek_cb(opaque, offset):
    print("SEEK", opaque, offset)

    stream_provider = ctypes.cast(opaque, ctypes.POINTER(ctypes.py_object)).contents.value

    stream_provider.seek(offset)

    return 0


@vlc.CallbackDecorators.MediaCloseCb
def media_close_cb(opaque):
    print("CLOSE", opaque)

    stream_provider = ctypes.cast(opaque, ctypes.POINTER(ctypes.py_object)).contents.value

    stream_provider.release_resources()


# MAIN
if __name__ == '__main__':
    parser = argparse.ArgumentParser(
            description='play files found in specified media folder (in alphabetic order)',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
            'media_folder',
            help='where to find files to play')
    parser.add_argument(
            '--extension',
            default='ts',
            help='file extension of the files to play')
    args = parser.parse_args()

    # helper object acting as media data provider
    # it is just to highlight how the opaque pointer in the callback can be used
    # and that the logic can be isolated from the callbacks
    stream_provider = StreamProviderDir(args.media_folder, args.extension)

    # these two lines to highlight how to pass a python object using ctypes
    # it is verbose, but you can see the steps required
    stream_provider_obj = ctypes.py_object(stream_provider)
    stream_provider_ptr = ctypes.byref(stream_provider_obj)

    # create an instance of vlc
    instance = vlc.Instance()

    # setup the callbacks for the media
    media = instance.media_new_callbacks(
            media_open_cb,
            media_read_cb,
            media_seek_cb,
            media_close_cb,
            stream_provider_ptr)
    player = media.player_new_from_media()

    # play/stop
    player.play()
    input("press enter to quit")
    player.stop()



#! /usr/bin/env python3

#
# gtk example/widget for VLC Python bindings
# Copyright (C) 2009-2010 the VideoLAN team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#

"""VLC Gtk Widget classes + example application.

This module provides two helper classes, to ease the embedding of a
VLC component inside a pygtk application.

VLCWidget is a simple VLC widget.

DecoratedVLCWidget provides simple player controls.

When called as an application, it behaves as a video player.

$Id$
"""

import gtk
gtk.gdk.threads_init()

import sys
import vlc

from gettext import gettext as _

# Create a single vlc.Instance() to be shared by (possible) multiple players.
instance = vlc.Instance()

class VLCWidget(gtk.DrawingArea):
    """Simple VLC widget.

    Its player can be controlled through the 'player' attribute, which
    is a vlc.MediaPlayer() instance.
    """
    def __init__(self, *p):
        gtk.DrawingArea.__init__(self)
        self.player = instance.media_player_new()
        def handle_embed(*args):
            if sys.platform == 'win32':
                self.player.set_hwnd(self.window.handle)
            else:
                self.player.set_xwindow(self.window.xid)
            return True
        self.connect("map", handle_embed)
        self.set_size_request(320, 200)

class DecoratedVLCWidget(gtk.VBox):
    """Decorated VLC widget.

    VLC widget decorated with a player control toolbar.

    Its player can be controlled through the 'player' attribute, which
    is a Player instance.
    """
    def __init__(self, *p):
        gtk.VBox.__init__(self)
        self._vlc_widget = VLCWidget(*p)
        self.player = self._vlc_widget.player
        self.pack_start(self._vlc_widget, expand=True)
        self._toolbar = self.get_player_control_toolbar()
        self.pack_start(self._toolbar, expand=False)

    def get_player_control_toolbar(self):
        """Return a player control toolbar
        """
        tb = gtk.Toolbar()
        tb.set_style(gtk.TOOLBAR_ICONS)
        for text, tooltip, stock, callback in (
            (_("Play"), _("Play"), gtk.STOCK_MEDIA_PLAY, lambda b: self.player.play()),
            (_("Pause"), _("Pause"), gtk.STOCK_MEDIA_PAUSE, lambda b: self.player.pause()),
            (_("Stop"), _("Stop"), gtk.STOCK_MEDIA_STOP, lambda b: self.player.stop()),
            ):
            b=gtk.ToolButton(stock)
            b.set_tooltip_text(tooltip)
            b.connect("clicked", callback)
            tb.insert(b, -1)
        tb.show_all()
        return tb

class VideoPlayer:
    """Example simple video player.
    """
    def __init__(self):
        self.vlc = DecoratedVLCWidget()

    def main(self, fname):
        self.vlc.player.set_media(instance.media_new(fname))
        w = gtk.Window()
        w.add(self.vlc)
        w.show_all()
        w.connect("destroy", gtk.main_quit)
        gtk.main()

class MultiVideoPlayer:
    """Example multi-video player.

    It plays multiple files side-by-side, with per-view and global controls.
    """
    def main(self, filenames):
        # Build main window
        window=gtk.Window()
        mainbox=gtk.VBox()
        videos=gtk.HBox()

        window.add(mainbox)
        mainbox.add(videos)

        # Create VLC widgets
        for fname in filenames:
            v = DecoratedVLCWidget()
            v.player.set_media(instance.media_new(fname))
            videos.add(v)

        # Create global toolbar
        tb = gtk.Toolbar()
        tb.set_style(gtk.TOOLBAR_ICONS)

        def execute(b, methodname):
            """Execute the given method on all VLC widgets.
            """
            for v in videos.get_children():
                getattr(v.player, methodname)()
            return True

        for text, tooltip, stock, callback, arg in (
            (_("Play"), _("Global play"), gtk.STOCK_MEDIA_PLAY, execute, "play"),
            (_("Pause"), _("Global pause"), gtk.STOCK_MEDIA_PAUSE, execute, "pause"),
            (_("Stop"), _("Global stop"), gtk.STOCK_MEDIA_STOP, execute, "stop"),
            ):
            b = gtk.ToolButton(stock)
            b.set_tooltip_text(tooltip)
            b.connect("clicked", callback, arg)
            tb.insert(b, -1)

        mainbox.pack_start(tb, expand=False)

        window.show_all()
        window.connect("destroy", gtk.main_quit)
        gtk.main()

if __name__ == '__main__':
    if not sys.argv[1:]:
       print('You must provide at least 1 movie filename')
       sys.exit(1)
    if len(sys.argv[1:]) == 1:
        # Only 1 file. Simple interface
        p=VideoPlayer()
        p.main(sys.argv[1])
    else:
        # Multiple files.
        p=MultiVideoPlayer()
        p.main(sys.argv[1:])
